import pytest

from xlviews.sheetframe import SheetFrame
from xlviews.statsframe import StatsGrouper


@pytest.fixture(scope="module")
def gr(sf_parent: SheetFrame):
    return StatsGrouper(sf_parent, ["x", "y"])


@pytest.mark.parametrize(("column", "c"), [("x", "C"), ("y", "D")])
def test_iter_formulas_list_index(gr: StatsGrouper, column, c):
    fs = list(gr.iter_formulas(column, ["min", "max"]))
    a = [f"=${c}${x}" for x in [4, 4, 8, 8, 12, 12, 16, 16]]
    assert fs == a


def test_iter_formulas_list_index_none(gr: StatsGrouper):
    fs = list(gr.iter_formulas("z", ["min", "max"]))
    assert fs == [""] * 8


@pytest.mark.parametrize(("column", "c"), [("a", "F"), ("b", "G"), ("c", "H")])
def test_iter_formulas_list_columns(gr: StatsGrouper, column, c):
    fs = list(gr.iter_formulas(column, ["min", "max"], wrap="__{}__"))
    assert len(fs) == 8
    assert fs[0] == f"=__AGGREGATE(5,7,${c}$4:${c}$7,${c}$20:${c}$23)__"
    assert fs[1] == f"=__AGGREGATE(4,7,${c}$4:${c}$7,${c}$20:${c}$23)__"
    assert fs[2] == f"=__AGGREGATE(5,7,${c}$8:${c}$11)__"
    assert fs[3] == f"=__AGGREGATE(4,7,${c}$8:${c}$11)__"
    assert fs[4] == f"=__AGGREGATE(5,7,${c}$12:${c}$15)__"
    assert fs[5] == f"=__AGGREGATE(4,7,${c}$12:${c}$15)__"
    assert fs[6] == f"=__AGGREGATE(5,7,${c}$16:${c}$19)__"
    assert fs[7] == f"=__AGGREGATE(4,7,${c}$16:${c}$19)__"


@pytest.mark.parametrize(("column", "c"), [("x", "C"), ("y", "D")])
def test_iter_formulas_dict_index(gr: StatsGrouper, column, c):
    fs = list(gr.iter_formulas(column, {}))
    a = [f"=${c}${x}" for x in [4, 8, 12, 16]]
    assert fs == a


def test_iter_formulas_dict_index_none(gr: StatsGrouper):
    fs = list(gr.iter_formulas("z", {}))
    assert fs == [""] * 4


@pytest.mark.parametrize(
    ("column", "c", "k"),
    [("a", "F", 1), ("b", "G", 12), ("c", "H", 9)],
)
def test_iter_formulas_dict_columns(gr: StatsGrouper, column, c, k):
    funcs = {"a": "mean", "b": "median", "c": "sum"}
    fs = list(gr.iter_formulas(column, funcs, wrap="__{}__"))
    assert len(fs) == 4
    x = f"=__AGGREGATE({k},7,${c}$4:${c}$7,${c}$20:${c}$23)__"
    assert fs[0] == x
    assert fs[1] == f"=__AGGREGATE({k},7,${c}$8:${c}$11)__"
    assert fs[2] == f"=__AGGREGATE({k},7,${c}$12:${c}$15)__"
    assert fs[3] == f"=__AGGREGATE({k},7,${c}$16:${c}$19)__"


def test_get_index(gr: StatsGrouper):
    index = gr.get_index(["a", "b"])
    assert index == ["a", "b", "a", "b", "a", "b", "a", "b"]


def test_get_columns_list(gr: StatsGrouper):
    columns = gr.get_columns([])
    assert columns == ["func", "x", "y", "z", "a", "b", "c"]


def test_get_columns_dict(gr: StatsGrouper):
    columns = gr.get_columns({"a": "mean", "b": "median"})
    assert columns == ["x", "y", "z", "a", "b", "c"]


@pytest.mark.parametrize(
    ("funcs", "shape"),
    [(["mean"], (4, 7)), (["min", "max", "median"], (12, 7)), ({"a": "count"}, (4, 6))],
)
def test_get_values_shape(gr: StatsGrouper, funcs, shape):
    assert gr.get_values(funcs).shape == shape


@pytest.mark.parametrize(
    ("funcs", "shape"),
    [(["mean"], (4, 3)), (["min", "max"], (8, 3)), ({"a": "std"}, (4, 3))],
)
def test_get_frame_shape(gr: StatsGrouper, funcs, shape):
    assert gr.get_frame(funcs).shape == shape


def test_get_frame_index_list(gr: StatsGrouper):
    df = gr.get_frame(["mean"])
    assert df.index.names == ["func", "x", "y", "z"]


def test_get_frame_index_dict(gr: StatsGrouper):
    df = gr.get_frame({})
    assert df.index.names == ["x", "y", "z"]


@pytest.mark.parametrize("funcs", [[], {}])
def test_get_frame_columns(gr: StatsGrouper, funcs):
    df = gr.get_frame(funcs)
    assert df.columns.to_list() == ["a", "b", "c"]


def test_get_frame_wrap_str(gr: StatsGrouper):
    df = gr.get_frame(["mean"], wrap="__{}__")
    values = df.to_numpy().flatten()
    assert all(x.startswith("=__AGGREGATE") for x in values)
    assert all(x.endswith(")__") for x in values)


def test_get_frame_wrap_dict(gr: StatsGrouper):
    df = gr.get_frame(["mean"], wrap={"a": "A{}A", "b": "B{}B"})
    assert all(x.startswith("=AAGGREGATE") for x in df["a"])
    assert all(x.endswith(")A") for x in df["a"])
    assert all(x.startswith("=BAGGREGATE") for x in df["b"])
    assert all(x.endswith(")B") for x in df["b"])
    assert all(x.startswith("=AGGREGATE") for x in df["c"])
    assert all(x.endswith(")") for x in df["c"])


def test_get_frame_offset(gr: StatsGrouper):
    df = gr.get_frame(["mean"]).reset_index()
    assert df["x"].iloc[0] == "=$C$4"
    assert df["y"].iloc[-1] == "=$D$16"
    assert df["a"].iloc[0] == "=AGGREGATE(1,7,$F$4:$F$7,$F$20:$F$23)"
    assert df["c"].iloc[-1] == "=AGGREGATE(1,7,$H$16:$H$19)"



import pytest
from pandas import DataFrame
from xlwings import Sheet

from xlviews.distframe import DistFrame
from xlviews.formula import NONCONST_VALUE
from xlviews.sheetframe import SheetFrame


@pytest.mark.parametrize(
    ("columns", "values"),
    [
        (None, ["x", "y", "a_n", "a_v", "a_s", "b_n", "b_v", "b_s"]),
        ("a", ["x", "y", "a_n", "a_v", "a_s"]),
        ("b", ["x", "y", "b_n", "b_v", "b_s"]),
    ],
)
def test_columns(df: DataFrame, sheet: Sheet, columns, values):
    sf = SheetFrame(3, 2, data=df, style=False, sheet=sheet)
    sf = DistFrame(sf, columns, by=["x", "y"])
    assert sf.columns == values


def test_group_error(sheet: Sheet):
    df = DataFrame(
        {
            "x": [1, 1, 1, 2, 2, 2, 1, 1, 1, 2, 2, 2, 2, 2],
            "y": [3, 3, 3, 3, 3, 4, 4, 4, 4, 3, 3, 3, 4, 4],
            "a": [5, 4, 3, 2, 1, 4, 3, 2, 1, 3, 2, 1, 2, 1],
            "b": [1, 2, 3, 4, 5, 1, 2, 3, 4, 1, 2, 3, 1, 2],
        },
    )
    df = df.set_index(["x", "y"])

    sf = SheetFrame(3, 2, data=df, style=False, sheet=sheet)
    with pytest.raises(ValueError, match="group must be continuous"):
        sf = DistFrame(sf, None, by=["x", "y"])


def test_group_const(sheet: Sheet):
    df = DataFrame(
        {
            "x": [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
            "y": [3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 4],
            "a": [5, 4, 3, 2, 1, 4, 3, 2, 1, 3, 2, 1, 2, 1],
            "b": [1, 2, 3, 4, 5, 1, 2, 3, 4, 1, 2, 3, 1, 2],
        },
    )
    df = df.set_index(["x", "y"])

    sf = SheetFrame(3, 2, data=df, style=False, sheet=sheet)
    sf = sf.distframe(by=["x", "y"])

    assert sf.sheet["H20"].value == 1
    assert sf.sheet["H21"].value == NONCONST_VALUE



import pytest
from pandas import DataFrame, MultiIndex
from xlwings import Sheet
from xlwings.constants import BordersIndex

from xlviews.config import rcParams
from xlviews.sheetframe import SheetFrame
from xlviews.utils import rgb


@pytest.mark.parametrize(("index", "value"), [("Vertical", 11), ("Horizontal", 12)])
def test_border_index(index, value):
    assert getattr(BordersIndex, f"xlInside{index}") == value


@pytest.mark.parametrize(
    ("weight", "value"),
    [((1, 2, 3, 4), (1, 2, -4138, 4)), (2, (2, 2, 2, 2))],
)
def test_border_edge(sheet: Sheet, weight, value):
    from xlviews.style import set_border_edge

    set_border_edge(sheet["C3:E5"], weight, color="red")
    assert sheet["B3:C5"].api.Borders(11).Weight == value[0]
    assert sheet["B3:C5"].api.Borders(11).Color == 255
    assert sheet["E3:F5"].api.Borders(11).Weight == value[1]
    assert sheet["E3:F5"].api.Borders(11).Color == 255
    assert sheet["C2:E3"].api.Borders(12).Weight == value[2]
    assert sheet["C2:E3"].api.Borders(12).Color == 255
    assert sheet["C5:E6"].api.Borders(12).Weight == value[3]
    assert sheet["C5:E6"].api.Borders(12).Color == 255


def test_border_inside(sheet: Sheet):
    from xlviews.style import set_border_inside

    set_border_inside(sheet["C3:E5"], weight=2, color="red")
    assert sheet["C3:E5"].api.Borders(11).Weight == 2
    assert sheet["C3:E5"].api.Borders(11).Color == 255
    assert sheet["C3:E5"].api.Borders(12).Weight == 2
    assert sheet["C3:E5"].api.Borders(12).Color == 255


def test_border(sheet: Sheet):
    from xlviews.style import set_border

    set_border(sheet["C3:E5"], edge_weight=2, inside_weight=1)
    assert sheet["B3:C5"].api.Borders(11).Weight == 2
    assert sheet["C3:E5"].api.Borders(11).Weight == 1


def test_border_zero(sheet: Sheet):
    from xlviews.style import set_border_line

    set_border_line(sheet["C3:D5"], "xlInsideVertical", weight=0, color="red")
    assert sheet["C3:D5"].api.Borders(11).Weight == 2
    assert sheet["C3:D5"].api.Borders(12).Weight == 2


def test_fill(sheet: Sheet):
    from xlviews.style import set_fill

    set_fill(sheet["C3:E5"], color="pink")
    assert sheet["C3:E5"].api.Interior.Color == 13353215


def test_font(sheet: Sheet):
    from xlviews.style import set_font

    rng = sheet["C3"]
    rng.value = "abc"
    set_font(rng, "Times", size=24, bold=True, italic=True, color="green")
    assert rng.api.Font.Name == "Times"
    assert rng.api.Font.Size == 24
    assert rng.api.Font.Bold == 1
    assert rng.api.Font.Italic == 1
    assert rng.api.Font.Color == 32768


def test_font_with_name(sheet: Sheet):
    from xlviews.config import rcParams
    from xlviews.style import set_font

    rng = sheet["C3"]
    rng.value = "abc"
    set_font(rng)
    assert rng.api.Font.Name == rcParams["chart.font.name"]


@pytest.mark.parametrize(
    ("align", "value"),
    [("right", -4152), ("left", -4131), ("center", -4108)],
)
def test_alignment_horizontal(sheet: Sheet, align, value):
    from xlviews.style import set_alignment

    rng = sheet["C3"]
    rng.value = "a"
    set_alignment(rng, horizontal_alignment=align)
    assert rng.api.HorizontalAlignment == value


@pytest.mark.parametrize(
    ("align", "value"),
    [("top", -4160), ("bottom", -4107), ("center", -4108)],
)
def test_alignment_vertical(sheet: Sheet, align, value):
    from xlviews.style import set_alignment

    rng = sheet["C3"]
    rng.value = "a"
    set_alignment(rng, vertical_alignment=align)
    assert rng.api.VerticalAlignment == value


@pytest.mark.parametrize(
    ("axis", "even_color", "odd_color"),
    [(0, 100, 200), (1, 300, 400)],
)
def test_banding(sheet: Sheet, axis, even_color, odd_color):
    from xlviews.style import set_banding

    rng = sheet["C3:F6"]
    set_banding(rng, axis, even_color, odd_color)
    assert rng.api.FormatConditions(1).Interior.Color == even_color
    assert rng.api.FormatConditions(2).Interior.Color == odd_color


def test_hide_succession(sheet: Sheet):
    from xlviews.style import hide_succession

    rng = sheet["C3:C8"]
    rng.options(transpose=True).value = [1, 1, 2, 2, 3, 3]
    rng = sheet["D3:D8"]
    rng.options(transpose=True).value = [1, 1, 1, 2, 2, 2]
    rng = sheet["C3:D8"]
    hide_succession(rng, color="red")
    assert rng.api.FormatConditions(1).Font.Color == 255


def test_hide_unique(sheet: Sheet):
    from xlviews.style import hide_unique

    rng = sheet["C3:C8"]
    rng.options(transpose=True).value = [1, 1, 2, 2, 3, 3]
    rng = sheet["D3:D8"]
    rng.options(transpose=True).value = [1, 1, 1, 1, 1, 1]
    rng = sheet["C2:D2"]
    rng.value = ["a", "b"]
    hide_unique(rng, 6, color="red")
    assert rng.api.FormatConditions(1).Font.Color == 255


def test_hide_gridlines(sheet: Sheet):
    from xlviews.style import hide_gridlines

    hide_gridlines(sheet)
    assert sheet.book.app.api.ActiveWindow.DisplayGridlines is False


@pytest.mark.parametrize(
    "name",
    ["index.name", "index", "columns.name", "columns", "values"],
)
@pytest.mark.parametrize("gray", [True, False])
def test_set_style(sheet: Sheet, name, gray):
    from xlviews.style import _set_style

    rng = sheet["C3:E5"]
    _set_style(rng[0], rng[-1], name, gray=gray)
    param = f"frame.{name}.fill.color"
    color = rgb("#eeeeee") if gray and name != "values" else rgb(rcParams[param])
    assert rng.api.Interior.Color == color


@pytest.fixture(scope="module")
def sf_basic(sheet_module: Sheet):
    from xlviews.style import set_frame_style

    df = DataFrame({"a": [1, 2, 3, 4], "b": [5, 6, 7, 8]})
    sf = SheetFrame(2, 2, data=df, style=False, sheet=sheet_module)
    set_frame_style(sf, autofit=True)
    return sf


@pytest.mark.parametrize(
    ("cell", "name"),
    [("B2", "index.name"), ("B6", "index"), ("D2", "columns"), ("D5", "values")],
)
def test_frame_style_basic(sf_basic: SheetFrame, cell: str, name: str):
    c = rgb(rcParams[f"frame.{name}.fill.color"])
    assert sf_basic.sheet[cell].api.Interior.Color == c


def test_frame_style_banding_succession(sheet_module: Sheet):
    from xlviews.style import set_frame_style

    df = DataFrame({"x": [1, 1, 2, 2], "a": [1, 2, 3, 4], "b": [5, 6, 7, 8]})
    df = df.set_index("x")
    sf = SheetFrame(2, 6, data=df, style=False, sheet=sheet_module)
    set_frame_style(sf, autofit=True, banding=True, succession=True)
    assert sf.sheet["F4"].api.FormatConditions(1)
    assert sf.sheet["H5"].api.FormatConditions(1)


@pytest.fixture(scope="module")
def df_mc():
    df_mc = DataFrame([[1, 2, 3], [4, 5, 6], [7, 8, 9], [10, 11, 12]])
    mi = [("a", "b", "c"), ("d", "e", "f"), ("g", "h", "i")]
    df_mc.columns = MultiIndex.from_tuples(mi)
    df_mc.columns.names = ["x", "y", "z"]
    return df_mc


@pytest.fixture(scope="module")
def sf_mc(sheet_module: Sheet, df_mc: DataFrame):
    from xlviews.style import set_frame_style

    sf = SheetFrame(9, 2, data=df_mc, style=False, sheet=sheet_module)
    set_frame_style(sf, autofit=True)
    return sf


@pytest.mark.parametrize(
    ("cell", "name"),
    [
        ("B9", "index.name"),
        ("B11", "index.name"),
        ("B12", "index"),
        ("B15", "index"),
        ("C9", "columns.name"),
        ("E10", "columns.name"),
        ("C11", "columns"),
        ("E11", "columns"),
        ("C12", "values"),
        ("E15", "values"),
    ],
)
def test_frame_style_multi_columns(sf_mc: SheetFrame, cell: str, name: str):
    c = rgb(rcParams[f"frame.{name}.fill.color"])
    assert sf_mc.sheet[cell].api.Interior.Color == c


@pytest.fixture(scope="module")
def df_mic(df_mc: DataFrame):
    df_mic = df_mc.copy()
    df_mic.columns = MultiIndex.from_tuples([("a", "b"), ("c", "d"), ("e", "f")])
    df_mic.columns.names = ["x", "y"]
    i = [("i", "j"), ("k", "l"), ("m", "n"), ("o", "p")]
    df_mic.index = MultiIndex.from_tuples(i)
    return df_mic


@pytest.fixture(scope="module")
def sf_mic(sheet_module: Sheet, df_mic: DataFrame):
    from xlviews.style import set_frame_style

    sf = SheetFrame(9, 7, data=df_mic, style=False, sheet=sheet_module)
    set_frame_style(sf, autofit=True)
    return sf


@pytest.mark.parametrize(
    ("cell", "name"),
    [
        ("G9", "index.name"),
        ("H10", "index.name"),
        ("G11", "index"),
        ("H14", "index"),
        ("I9", "columns.name"),
        ("K9", "columns.name"),
        ("I10", "columns"),
        ("K10", "columns"),
        ("I11", "values"),
        ("K14", "values"),
    ],
)
def test_frame_style_multi_index_columns(sf_mic: SheetFrame, cell: str, name: str):
    c = rgb(rcParams[f"frame.{name}.fill.color"])
    assert sf_mic.sheet[cell].api.Interior.Color == c


@pytest.fixture(scope="module")
def sf_wide(sheet_module: Sheet):
    from xlviews.style import set_wide_column_style

    df = DataFrame({"x": ["i", "j"], "y": ["k", "l"], "a": [1, 2], "b": [3, 4]})
    sf = SheetFrame(24, 2, data=df, style=False, sheet=sheet_module)
    sf.add_wide_column("u", range(3), autofit=True)
    sf.add_wide_column("v", range(4), autofit=True)
    set_wide_column_style(sf, gray=False)
    return sf


@pytest.mark.parametrize(
    ("cell", "name"),
    [("G23", ".name"), ("M23", ".name"), ("G24", ""), ("M24", "")],
)
def test_frame_style_wide(sf_wide: SheetFrame, cell: str, name: str):
    c = rgb(rcParams[f"frame.wide-columns{name}.fill.color"])
    assert sf_wide.sheet[cell].api.Interior.Color == c


def test_table_style(sheet_module: Sheet):
    from xlviews.style import set_table_style

    df = DataFrame({"x": [1, 1, 2, 2], "a": [1, 2, 3, 4], "b": [5, 6, 7, 8]})
    df = df.set_index("x")
    sf = SheetFrame(17, 2, data=df, style=False, sheet=sheet_module)
    table = sf.as_table(style=False)
    set_table_style(table)
    assert table.sheet.book.api.TableStyles("xlviews")



if __name__ == "__main__":
    import xlwings as xw
    from xlwings.constants import ChartType

    from xlviews.axes import Axes
    from xlviews.decorators import quit_apps

    quit_apps()
    book = xw.Book()
    sheet_module = book.sheets.add()

    ct = ChartType.xlXYScatterLines
    ax = Axes(300, 10, chart_type=ct, sheet=sheet_module)
    x = sheet_module["B2:B11"]
    y = sheet_module["C2:C11"]
    x.options(transpose=True).value = list(range(10))
    y.options(transpose=True).value = list(range(10, 20))

    s = ax.add_series(x, y, label="a")



from __future__ import annotations

from itertools import product
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
from pandas import DataFrame

from xlviews.config import rcParams
from xlviews.decorators import turn_off_screen_updating
from xlviews.sheetframe import SheetFrame
from xlviews.style import set_alignment, set_font
from xlviews.utils import iter_columns

if TYPE_CHECKING:
    from xlwings import Range


class DistFrame(SheetFrame):
    parent: SheetFrame
    dist_func: dict[str, str]

    @turn_off_screen_updating
    def __init__(
        self,
        parent: SheetFrame,
        columns: str | list[str] | None = None,
        *,
        by: str | list[str] | None = None,
        dist: str | dict[str, str] = "norm",
        style: bool = True,
        gray: bool = True,
        autofit: bool = True,
        **kwargs,
    ) -> None:
        if columns is None:
            columns = parent.value_columns
        elif isinstance(columns, str):
            columns = [columns]

        self.dist_func = get_dist_func(dist, columns)

        if not parent.table:
            parent.as_table()

        by = list(iter_columns(parent, by)) if by else None
        data = get_init_data(parent, columns, by)

        super().__init__(
            data=data,
            parent=parent,
            index=by is not None,
            style=False,
            **kwargs,
        )

        if by:
            self.link_to_index(by)

        grouped = self.groupby(by)

        for column in columns:
            dist = self.dist_func[column]
            parent_column = self.parent.range(column).column
            column_int = self.range(column + "_n").column

            for row in grouped.values():
                if len(row) != 1:
                    raise ValueError("group must be continuous")

                start = row[0][0]
                length = row[0][1] - start + 1

                parent_cell = self.parent.sheet.range(start, parent_column)
                cell = self.sheet.range(start, column_int)

                formula = counter(parent_cell)
                set_formula(cell, length, formula)

                formula = sorted_value(parent_cell, cell, length)
                set_formula(cell.offset(0, 1), length, formula)

                formula = sigma_value(cell, length, dist)
                set_formula(cell.offset(0, 2), length, formula)

        for column in columns:
            fmt = self.parent.range(column).number_format
            self.set_number_format({f"{column}_v": fmt, f"{column}_s": "0.00"})

        if style:
            self.set_style(autofit=autofit, gray=gray)

        self.const_values()

    def link_to_index(self, by: list[str]) -> None:
        start = self.row + 1
        end = start + len(self) - 1
        for column in by:
            ref = self.parent.index(column)
            ref = self.parent.sheet.range(start, ref)
            ref = ref.get_address(row_absolute=False)
            formula = f"={ref}"
            to = self.index(column)
            range_ = self.sheet.range((start, to), (end, to))
            range_.value = formula

    def const_values(self) -> None:
        index = self.parent.index_columns
        array = np.zeros((len(index), 1))
        df = DataFrame(array, columns=["value"], index=index)
        sf = SheetFrame(data=df, head=self, gray=True, autofit=False)
        head = self.parent.cell.offset(-1, 0)
        tail = sf.cell.offset(1, 1)
        for k in range(len(index)):
            formula = "=" + head.offset(0, k).get_address()
            tail.offset(k, 0).value = formula

    def plot(
        self,
        x,
        label="auto",
        color=None,
        marker=None,
        axes=None,
        xlabel="auto",
        ylabel="auto",
        **kwargs,
    ):
        if ylabel == "auto":
            dist = self.dist_func[x] if isinstance(x, str) else self.dist_func[x[0]]
            ylabel = "σ" if dist == "norm" else "ln(-ln(1-F))"
        plot = None
        if isinstance(x, str) and xlabel == "auto":
            x_ = x.split("_")[0]
            xlabel = rcParams.get(f"axis.label.{x_}", x)
            if "_" in x and "[" in xlabel:
                xlabel = x + " " + xlabel[xlabel.index("[") :]
        xs = [x] if isinstance(x, str) else x
        colors = color if isinstance(color, list) else [color] * len(xs)
        markers = marker if isinstance(marker, list) else [marker] * len(xs)
        for x_, color, marker in zip(xs, colors, markers, strict=False):
            label_ = x_ if label == "auto" and isinstance(x, list) else label
            plot = self._plot(
                x_,
                label=label_,
                axes=axes,
                color=color,
                marker=marker,
                xlabel=xlabel,
                ylabel=ylabel,
                **kwargs,
            )
            axes = plot.axes
        return plot

    def _plot(self, x, **kwargs):
        plot = super().plot(f"{x}_v", f"{x}_s", yformat="0_ ", **kwargs)
        # if fit:
        #     sigma = 2 if fit is True else fit
        #     column = self.add_column_for_fit(x, sigma)
        #     kwargs['marker'] = None
        #     kwargs['line'] = None
        #     kwargs.pop('axes')
        #     kwargs.pop('label')
        #     plot_ = super().plot(f'{x}_v', column, axes=plot.axes,
        #                          label=None, **kwargs)
        #     for series in plot_.series_collection:
        #         trendline = series.Trendlines().Add()
        #         plot_.axes.labels.append('__trendline__')
        #         trendline.DisplayEquation = True
        #         # trendline.Forward = 10
        #         # trendline.Backward = 10
        #     # print(plot_.axes.labels)
        #     # print(plot_.axes.legend.LegendEntries())
        #     # print(plot_.legend)
        #     plot_.axes.set_legend(**plot_.legend)
        return plot

    def fit(self, x):
        pass

    def add_column_for_fit(self, x, sigma):
        """

        Parameters
        ----------
        x : str
            変数名
        sigma: int or float
            フィッティングに用いるσ値の範囲
        """
        column_ = f"{x}_sf"
        if column_ in self.columns:
            return column_
        self[column_] = 1
        column = self.index(column_)
        sigma_cell = self.sheet.range(self.row - 1, column)
        sigma_cell.value = sigma
        set_font(sigma_cell, size=8, bold=True, italic=True, color="green")
        set_alignment(sigma_cell, "center")
        sigma = sigma_cell.get_address()
        row = self.cell.offset(self.columns_level).row
        column_ref = self.index(f"{x}_s")
        cell_ref = self.sheet.range(row, column_ref)
        cell_ref = cell_ref.get_address(row_absolute=False)
        cell = self.sheet.range(row, column)
        range_ = self.sheet.range(cell, cell.offset(len(self) - 1))
        range_.api.NumberFormatLocal = "0.00_ "
        formula = f"=IF(AND({cell_ref}>=-{sigma},{cell_ref}<={sigma}),{cell_ref},NA())"
        range_.value = formula
        return column_


def get_init_data(
    sf: SheetFrame,
    columns: list[str],
    by: list[str] | None,
) -> DataFrame:
    columns = [f"{c}_{suffix}" for c, suffix in product(columns, ["n", "v", "s"])]

    array = np.zeros((len(sf), len(columns)))
    df = DataFrame(array, columns=columns)

    if by:
        index = np.zeros((len(sf), len(by)))
        index = DataFrame(index, columns=by)
        df = pd.concat([index, df], axis=1)
        df = df.set_index(by)

    return df


def get_dist_func(dist: str | dict[str, str], columns: list[str]) -> dict[str, str]:
    if isinstance(dist, str):
        return {column: dist for column in columns}

    dist = dist.copy()
    for column in columns:
        dist.setdefault(column, "norm")

    return dist


def counter(cell: Range) -> str:
    start = cell.get_address()
    end = cell.get_address(row_absolute=False)
    return f"=AGGREGATE(3,1,{start}:{end})"


def sorted_value(parent_cell: Range, cell: Range, length: int) -> str:
    start = parent_cell.get_address()
    end = parent_cell.offset(length - 1).get_address()
    small = cell.get_address(row_absolute=False)
    return f"=IF({small}>0,AGGREGATE(15,1,{start}:{end},{small}),NA())"


def sigma_value(cell: Range, length: int, dist: str) -> str:
    small = cell.get_address(row_absolute=False)
    end = cell.offset(length - 1).get_address()

    if dist == "norm":
        return f"=IF({small}>0,NORM.S.INV({small}/({end}+1)),NA())"

    if dist == "weibull":
        return f"=IF({small}>0,LN(-LN(1-{small}/({end}+1))),NA())"

    raise ValueError("不明な分布", dist)


def set_formula(cell: Range, length: int, formula: str) -> None:
    end = cell.offset(length - 1)
    cell.sheet.range(cell, end).value = formula



from xlviews.axes import Axes
from xlviews.distframe import DistFrame
from xlviews.grouper import Grouper
from xlviews.series import Series
from xlviews.sheetframe import SheetFrame
from xlviews.statsframe import StatsFrame
from xlviews.utils import constant

__all__ = [
    "Axes",
    "DistFrame",
    "Grouper",
    "Series",
    "SheetFrame",
    "StatsFrame",
    "constant",
]



from __future__ import annotations

from functools import wraps
from typing import TYPE_CHECKING

import xlwings as xw

if TYPE_CHECKING:
    from collections.abc import Callable


def turn_off_screen_updating(func: Callable) -> Callable:
    """Turn screen updating off to speed up your script."""

    @wraps(func)
    def _func(*args, **kwargs):  # noqa: ANN202
        if app := xw.apps.active:
            is_updating = app.screen_updating
            app.screen_updating = False

        try:
            return func(*args, **kwargs)
        finally:
            if app:
                app.screen_updating = is_updating

    return _func



import numpy as np
import pytest

from xlviews.sheetframe import SheetFrame
from xlviews.statsframe import StatsFrame


@pytest.fixture(scope="module", params=[True, False])
def sf(sf_parent: SheetFrame, request: pytest.FixtureRequest):
    funcs = {"a": "max", "b": "std", "c": "mean"}
    return StatsFrame(sf_parent, funcs, by=":y", table=request.param)


def test_len(sf: SheetFrame):
    assert len(sf) == 4


def test_columns(sf: SheetFrame):
    assert sf.columns == ["x", "y", "z", "a", "b", "c"]


def test_index_columns(sf: SheetFrame):
    assert sf.index_columns == ["x", "y", "z"]


def test_value_columns(sf: SheetFrame):
    assert sf.value_columns == ["a", "b", "c"]


@pytest.mark.parametrize(
    ("cell", "value"),
    [
        ("C3:C7", ["x", "a", "a", "b", "b"]),
        ("D3:D7", ["y", "c", "d", "c", "d"]),
        ("E3:E7", ["z", None, None, None, None]),
        ("F3:F7", ["a", 18, 7, 11, 15]),
        ("H3:H7", ["c", 18, (28 + 30 + 34) / 3, 38 / 3, 7]),
    ],
)
def test_value(sf: SheetFrame, cell, value):
    assert sf.sheet[cell].value == value


@pytest.mark.parametrize(
    ("cell", "value"),
    [
        ("G4", np.std([0, 1, 2, 3, 18, 21, 24, 27])),
        ("G5", np.std([4, 5, 6, 7])),
        ("G6", np.std([8, 9, 0, 3])),
        ("G7", np.std([6, 9, 12, 15])),
    ],
)
def test_value_std(sf: SheetFrame, cell, value):
    v = sf.sheet[cell].value
    assert v is not None
    np.testing.assert_allclose(v, value)



import numpy as np
import pytest
from pandas import DataFrame, MultiIndex, Series
from xlwings import Sheet

from xlviews.sheetframe import SheetFrame


@pytest.fixture(scope="module")
def df():
    df = DataFrame(
        {
            "x": [1, 1, 1, 1, 2, 2, 2, 2],
            "y": [1, 1, 2, 2, 1, 1, 2, 2],
            "a": [1, 2, 3, 4, 5, 6, 7, 8],
            "b": [11, 12, 13, 14, 15, 16, 17, 18],
        },
    )
    return df.set_index(["x", "y"])


def test_df(df: DataFrame):
    assert len(df) == 8
    assert df.shape == (8, 2)
    assert df.columns.to_list() == ["a", "b"]
    x = [(1, 1), (1, 1), (1, 2), (1, 2), (2, 1), (2, 1), (2, 2), (2, 2)]
    assert df.index.to_list() == x
    assert df.index.names == ["x", "y"]
    assert isinstance(df.index, MultiIndex)


@pytest.fixture(scope="module")
def sf(df: DataFrame, sheet_module: Sheet):
    return SheetFrame(10, 6, data=df, style=False, sheet=sheet_module)


def test_value(sf: SheetFrame):
    v = sf.cell.expand().options(ndim=2).value
    assert len(v) == 9
    assert v[0] == ["x", "y", "a", "b"]
    assert v[1] == [1, 1, 1, 11]
    assert v[2] == [1, 1, 2, 12]
    assert v[3] == [1, 2, 3, 13]
    assert v[4] == [1, 2, 4, 14]
    assert v[5] == [2, 1, 5, 15]
    assert v[6] == [2, 1, 6, 16]
    assert v[7] == [2, 2, 7, 17]
    assert v[8] == [2, 2, 8, 18]


def test_init(sf: SheetFrame, sheet_module: Sheet):
    assert sf.row == 10
    assert sf.column == 6
    assert sf.sheet.name == sheet_module.name
    assert sf.index_level == 2
    assert sf.columns_level == 1


def test_set_data_from_sheet(sf: SheetFrame):
    sf.set_data_from_sheet(index_level=0)
    assert sf.has_index is False
    assert sf.index_columns == []
    assert sf.value_columns == ["x", "y", "a", "b"]
    sf.set_data_from_sheet(index_level=2)
    assert sf.has_index is True
    assert sf.index_columns == ["x", "y"]
    assert sf.value_columns == ["a", "b"]


def test_init_index_false(df: DataFrame, sheet: Sheet):
    sf = SheetFrame(2, 3, data=df, index=False, style=False, sheet=sheet)
    assert sf.columns == ["a", "b"]
    assert sf.index_level == 0


def test_len(sf: SheetFrame):
    assert len(sf) == 8


def test_columns(sf: SheetFrame):
    assert sf.columns == ["x", "y", "a", "b"]


def test_value_columns(sf: SheetFrame):
    assert sf.value_columns == ["a", "b"]


def test_index_columns(sf: SheetFrame):
    assert sf.index_columns == ["x", "y"]


def test_contains(sf: SheetFrame):
    assert "x" in sf
    assert "a" in sf


def test_iter(sf: SheetFrame):
    assert list(sf) == ["x", "y", "a", "b"]


@pytest.mark.parametrize(
    ("column", "relative", "index"),
    [
        ("a", True, 3),
        ("a", False, 8),
        ("b", True, 4),
        ("b", False, 9),
        (["x", "b"], True, [1, 4]),
        (["y", "b"], False, [7, 9]),
    ],
)
def test_index(sf: SheetFrame, column, relative, index):
    assert sf.index(column, relative=relative) == index


def test_index_error(sf: SheetFrame):
    with pytest.raises(ValueError, match="'z' is not in list"):
        sf.index("z")


def test_data(sf: SheetFrame, df: DataFrame):
    df_ = sf.data
    np.testing.assert_array_equal(df_.index, df.index)
    np.testing.assert_array_equal(df_.index.names, df.index.names)
    np.testing.assert_array_equal(df_.columns, df.columns)
    np.testing.assert_array_equal(df_.columns.names, df.columns.names)
    np.testing.assert_array_equal(df_, df)
    assert df_.index.name == df.index.name
    assert df_.columns.name == df.columns.name


def test_range_all(sf: SheetFrame):
    assert sf.range_all().get_address() == "$F$10:$I$18"
    assert sf.range().get_address() == "$F$10:$I$18"


@pytest.mark.parametrize(
    ("start", "end", "address"),
    [
        (False, None, "$F$10:$G$18"),
        (-1, None, "$F$10:$G$10"),
        (0, None, "$F$11:$G$11"),
        (None, None, "$F$11:$G$18"),
        (20, None, "$F$20:$G$20"),
        (20, 100, "$F$20:$G$100"),
    ],
)
def test_range_index(sf: SheetFrame, start, end, address):
    assert sf.range_index(start, end).get_address() == address
    assert sf.range("index", start, end).get_address() == address


@pytest.mark.parametrize(
    ("column", "start", "end", "address"),
    [
        ("x", -1, None, "$F$10"),
        ("y", -1, None, "$G$10"),
        ("a", -1, None, "$H$10"),
        ("b", -1, None, "$I$10"),
        ("x", 1, None, "$F$1"),
        ("a", 2, None, "$H$2"),
        ("b", 100, None, "$I$100"),
        ("y", None, None, "$G$11:$G$18"),
        ("a", False, None, "$H$10:$H$18"),
        ("b", False, None, "$I$10:$I$18"),
        ("x", 2, 100, "$F$2:$F$100"),
        ("a", 3, 300, "$H$3:$H$300"),
    ],
)
def test_range_column(sf: SheetFrame, column, start, end, address):
    assert sf.range_column(column, start, end).get_address() == address
    assert sf.range(column, start, end).get_address() == address


def test_getitem_str(sf: SheetFrame):
    s = sf["a"]
    assert isinstance(s, Series)
    assert s.name == "a"
    np.testing.assert_array_equal(s, range(1, 9))


def test_getitem_list(sf: SheetFrame):
    df = sf[["x", "a"]]
    assert isinstance(df, DataFrame)
    assert df.columns.to_list() == ["x", "a"]
    x = np.array([[1, 1, 1, 1, 2, 2, 2, 2], range(1, 9)]).T
    np.testing.assert_array_equal(df, x)


@pytest.mark.parametrize(
    ("by", "v1", "v2"),
    [
        ("x", [(11, 14)], [(15, 18)]),
        ("y", [(11, 12), (15, 16)], [(13, 14), (17, 18)]),
    ],
)
def test_groupby(sf: SheetFrame, by, v1, v2):
    g = sf.groupby(by)
    assert len(g) == 2
    assert g[(1,)] == v1
    assert g[(2,)] == v2


def test_groupby_list(sf: SheetFrame):
    g = sf.groupby(["x", "y"])
    assert len(g) == 4
    assert g[(1, 1)] == [(11, 12)]
    assert g[(1, 2)] == [(13, 14)]
    assert g[(2, 1)] == [(15, 16)]
    assert g[(2, 2)] == [(17, 18)]



import numpy as np
import pytest
from pandas import DataFrame, Series
from xlwings import Sheet

from xlviews.sheetframe import SheetFrame


@pytest.fixture(scope="module")
def df():
    df = DataFrame({"x": ["i", "j"], "y": ["k", "l"], "a": [1, 2], "b": [3, 4]})
    return df.set_index(["x", "y"])


def test_df(df: DataFrame):
    assert len(df) == 2
    assert df.shape == (2, 2)
    assert df.columns.to_list() == ["a", "b"]
    assert df.index.to_list() == [("i", "k"), ("j", "l")]
    assert df.index.names == ["x", "y"]


@pytest.fixture(scope="module")
def sf(df: DataFrame, sheet_module: Sheet):
    sf = SheetFrame(4, 2, data=df, style=False, sheet=sheet_module)
    sf.add_wide_column("u", range(3), autofit=False)
    sf.add_wide_column("v", range(4), autofit=False)
    return sf


def test_value(sf: SheetFrame):
    v = sf.cell.expand().options(ndim=2).value
    assert len(v) == 3
    assert v[0] == ["x", "y", "a", "b", *range(3), *range(4)]
    assert v[1] == ["i", "k", 1, 3, *([None] * 7)]
    assert v[2] == ["j", "l", 2, 4, *([None] * 7)]

    assert sf.cell.offset(-1, 4).value == "u"
    assert sf.cell.offset(-1, 7).value == "v"


def test_init(sf: SheetFrame, sheet_module: Sheet):
    assert sf.row == 4
    assert sf.column == 2
    assert sf.sheet.name == sheet_module.name
    assert sf.index_level == 2
    assert sf.columns_level == 1


def test_len(sf: SheetFrame):
    assert len(sf) == 2


def test_columns(sf: SheetFrame):
    assert sf.columns == ["x", "y", "a", "b", *range(3), *range(4)]


def test_value_columns(sf: SheetFrame):
    assert sf.value_columns == ["a", "b", *range(3), *range(4)]


def test_index_columns(sf: SheetFrame):
    assert sf.index_columns == ["x", "y"]


def test_wide_columns(sf: SheetFrame):
    assert sf.wide_columns == ["u", "v"]


def test_contains(sf: SheetFrame):
    assert "x" in sf
    assert "a" in sf
    assert "u" not in sf


def test_iter(sf: SheetFrame):
    assert list(sf) == ["x", "y", "a", "b", *range(3), *range(4)]


@pytest.mark.parametrize(
    ("column", "relative", "index"),
    [
        ("a", True, 3),
        ("a", False, 4),
        ("b", True, 4),
        ("b", False, 5),
        (["x", "b"], True, [1, 4]),
        (["y", "b"], False, [3, 5]),
    ],
)
def test_index(sf: SheetFrame, column, relative, index):
    assert sf.index(column, relative=relative) == index


@pytest.mark.parametrize(
    ("column", "relative", "index"),
    [
        ("u", True, (5, 7)),
        ("v", True, (8, 11)),
        ("u", False, (6, 8)),
        ("v", False, (9, 12)),
        (("u", 0), True, 5),
        (("u", 2), True, 7),
        (("v", 0), False, 9),
        (("v", 3), False, 12),
    ],
)
def test_index_wide(sf: SheetFrame, column, relative, index):
    assert sf.index_wide(column, relative=relative) == index


@pytest.mark.parametrize("column", ["z", ("u", -1)])
def test_index_error(sf: SheetFrame, column):
    with pytest.raises(ValueError, match=".* is not in list"):
        sf.index(column)


def test_data(sf: SheetFrame, df: DataFrame):
    df_ = sf.data
    np.testing.assert_array_equal(df_.index, df.index)
    np.testing.assert_array_equal(df_.index.names, df.index.names)
    np.testing.assert_array_equal(df_.columns[:2], df.columns)
    np.testing.assert_array_equal(df_.columns[2:], [*range(3), *range(4)])
    np.testing.assert_array_equal(df_.columns.names, df.columns.names)
    np.testing.assert_array_equal(df_.iloc[:, :2], df)
    assert df_.iloc[:, 2:].isna().all().all()
    assert df_.index.name == df.index.name
    assert df_.columns.name == df.columns.name


def test_range_all(sf: SheetFrame):
    assert sf.range_all().get_address() == "$B$4:$L$6"
    assert sf.range().get_address() == "$B$4:$L$6"


@pytest.mark.parametrize(
    ("start", "end", "address"),
    [
        (False, None, "$B$4:$C$6"),
        (-1, None, "$B$4:$C$4"),
        (0, None, "$B$5:$C$5"),
        (None, None, "$B$5:$C$6"),
        (20, None, "$B$20:$C$20"),
        (20, 100, "$B$20:$C$100"),
    ],
)
def test_range_index(sf: SheetFrame, start, end, address):
    assert sf.range_index(start, end).get_address() == address
    assert sf.range("index", start, end).get_address() == address


@pytest.mark.parametrize(
    ("column", "start", "end", "address"),
    [
        ("b", False, None, "$E$4:$E$6"),
        ("x", -1, None, "$B$4"),
        ("y", 0, None, "$C$5"),
        ("a", None, None, "$D$5:$D$6"),
        (("u", 0), -1, None, "$F$3:$F$4"),
        (("u", 2), 0, None, "$H$5"),
        (("v", 0), None, None, "$I$5:$I$6"),
        (("v", 3), False, None, "$L$3:$L$6"),
        ("u", -1, None, "$F$3:$H$4"),
        ("u", 0, None, "$F$5:$H$5"),
        ("v", None, None, "$I$5:$L$6"),
        ("v", False, None, "$I$3:$L$6"),
    ],
)
def test_range_column(sf: SheetFrame, column, start, end, address):
    assert sf.range_column(column, start, end).get_address() == address
    assert sf.range(column, start, end).get_address() == address


@pytest.mark.parametrize(
    ("column", "value"),
    [("a", [1, 2]), (("u", 1), [np.nan, np.nan])],
)
def test_getitem_str(sf: SheetFrame, column, value):
    s = sf[column]
    assert isinstance(s, Series)
    assert s.name == column
    np.testing.assert_array_equal(s, value)


def test_getitem_list(sf: SheetFrame):
    df = sf[["b", ("v", 3)]]
    assert isinstance(df, DataFrame)
    assert df.columns.to_list() == ["b", ("v", 3)]
    x = [[3, np.nan], [4, np.nan]]
    np.testing.assert_array_equal(df, x)



"""Set styles such as Marker."""

from __future__ import annotations

import itertools
from functools import partial
from typing import TYPE_CHECKING

import pywintypes
import seaborn as sns
from xlwings import Range, Sheet
from xlwings.constants import (
    BordersIndex,
    FormatConditionType,
    LineStyle,
    MarkerStyle,
    ScaleType,
    TableStyleElementType,
)

from xlviews.config import rcParams
from xlviews.decorators import turn_off_screen_updating
from xlviews.range import reference
from xlviews.utils import constant, rgb

if TYPE_CHECKING:
    from xlwings._xlwindows import COMRetryObjectWrapper

    from xlviews.sheetframe import SheetFrame
    from xlviews.table import Table


def set_border_line(
    rng: Range,
    index: str,
    weight: int = 2,
    color: int | str = 0,
) -> None:
    if not weight:
        return

    borders = rng.api.Borders
    border = borders(getattr(BordersIndex, index))
    border.LineStyle = LineStyle.xlContinuous
    border.Weight = weight
    border.Color = rgb(color)


def set_border_edge(
    rng: Range,
    weight: int | tuple[int, int, int, int] = 3,
    color: int | str = 0,
) -> None:
    if isinstance(weight, int):
        wl = wr = wt = wb = weight
    else:
        wl, wr, wt, wb = weight

    sheet = rng.sheet
    start, end = rng[0], rng[-1]

    left = sheet.range((start.row, start.column - 1), (end.row, start.column))
    set_border_line(left, "xlInsideVertical", weight=wl, color=color)

    right = sheet.range((start.row, end.column), (end.row, end.column + 1))
    set_border_line(right, "xlInsideVertical", weight=wr, color=color)

    top = sheet.range((start.row - 1, start.column), (start.row, end.column))
    set_border_line(top, "xlInsideHorizontal", weight=wt, color=color)

    bottom = sheet.range((end.row, start.column), (end.row + 1, end.column))
    set_border_line(bottom, "xlInsideHorizontal", weight=wb, color=color)


def set_border_inside(rng: Range, weight: int = 1, color: int | str = 0) -> None:
    set_border_line(rng, "xlInsideVertical", weight=weight, color=color)
    set_border_line(rng, "xlInsideHorizontal", weight=weight, color=color)


def set_border(
    rng: Range,
    edge_weight: int | tuple[int, int, int, int] = 2,
    inside_weight: int = 1,
    edge_color: int | str = 0,
    inside_color: int | str = rgb(140, 140, 140),
) -> None:
    if edge_weight:
        set_border_edge(rng, edge_weight, edge_color)

    if inside_weight:
        set_border_inside(rng, inside_weight, inside_color)


def set_fill(rng: Range, color: int | str | None = None) -> None:
    if color is not None:
        rng.api.Interior.Color = rgb(color)


def set_font_api(
    api: COMRetryObjectWrapper,
    name: str | None = None,
    *,
    size: float | None = None,
    bold: bool | None = None,
    italic: bool | None = None,
    color: int | str | None = None,
) -> None:
    name = name or rcParams["chart.font.name"]

    font = api.Font
    font.Name = name  # type: ignore
    if size:
        font.Size = size  # type: ignore
    if bold is not None:
        font.Bold = bold  # type: ignore
    if italic is not None:
        font.Italic = italic  # type: ignore
    if color is not None:
        font.Color = rgb(color)  # type: ignore


def set_font(rng: Range, *args, **kwargs) -> None:
    set_font_api(rng.api, *args, **kwargs)


def set_alignment(
    rng: Range,
    horizontal_alignment: str | None = None,
    vertical_alignment: str | None = None,
) -> None:
    if horizontal_alignment:
        rng.api.HorizontalAlignment = constant(horizontal_alignment)

    if vertical_alignment:
        rng.api.VerticalAlignment = constant(vertical_alignment)


def set_banding(
    rng: Range,
    axis: int = 0,
    even_color: int | str = rgb(240, 250, 255),
    odd_color: int | str = rgb(255, 255, 255),
) -> None:
    def banding(mod: int, color: int) -> None:
        formula = f"=MOD(ROW(), 2)={mod}" if axis == 0 else f"=MOD(COLUMN(), 2)={mod}"
        condition = add(Type=FormatConditionType.xlExpression, Formula1=formula)

        condition.SetFirstPriority()
        condition.StopIfTrue = False

        interior = condition.Interior
        interior.PatternColorIndex = constant("automatic")
        interior.Color = color
        interior.TintAndShade = 0

    add = rng.api.FormatConditions.Add

    banding(0, rgb(odd_color))
    banding(1, rgb(even_color))


def hide_succession(rng: Range, color: int | str = rgb(200, 200, 200)) -> None:
    cell = rng[0].get_address(row_absolute=False, column_absolute=False)

    start = rng[0].offset(-2).get_address(column_absolute=False)
    column = rng[0].offset(-1)
    column = ":".join(
        [
            column.get_address(column_absolute=False),
            column.get_address(row_absolute=False, column_absolute=False),
        ],
    )

    ref = (
        f"INDIRECT(ADDRESS(MAX(INDEX(SUBTOTAL(3,OFFSET({start},"
        f'ROW(INDIRECT("1:"&ROWS({column}))),))*ROW({column}),)),'
        f"COLUMN({column})))"
    )
    formula = f"={cell}={ref}"

    add = rng.api.FormatConditions.Add
    condition = add(Type=FormatConditionType.xlExpression, Formula1=formula)
    condition.SetFirstPriority()
    condition.StopIfTrue = False
    condition.Font.Color = rgb(color)


def hide_unique(rng: Range, length: int, color: int | str = rgb(100, 100, 100)) -> None:
    def address(r: Range) -> str:
        return r.get_address(row_absolute=False, column_absolute=False)

    start = rng[0, 0].offset(1, 0)
    end = rng[0, 0].offset(length, 0)
    cell = address(Range(start, end))
    ref = address(start)
    formula = f"=COUNTIF({cell}, {ref}) = {length}"

    add = rng.api.FormatConditions.Add
    condition = add(Type=FormatConditionType.xlExpression, Formula1=formula)
    condition.SetFirstPriority()
    condition.StopIfTrue = False
    condition.Font.Color = rgb(color)
    condition.Font.Italic = True


def hide_gridlines(sheet: Sheet) -> None:
    sheet.book.app.api.ActiveWindow.DisplayGridlines = False


def _set_style(
    start: Range,
    end: Range,
    name: str,
    *,
    border: bool = True,
    gray: bool = False,
    font: bool = True,
    fill: bool = True,
    font_size: int | None = None,
) -> None:
    rng = start.sheet.range(start, end)

    if border:
        set_border(rng, edge_color=rcParams["frame.gray.border.color"] if gray else 0)

    if fill:
        _set_style_fill(rng, name, gray=gray)

    if font:
        _set_style_font(rng, name, gray=gray, font_size=font_size)


def _set_style_fill(rng: Range, name: str, *, gray: bool = False) -> None:
    if gray and name != "values":
        color = rcParams["frame.gray.fill.color"]
    else:
        color = rcParams[f"frame.{name}.fill.color"]

    set_fill(rng, color=color)


def _set_style_font(
    rng: Range,
    name: str,
    *,
    gray: bool = False,
    font_size: int | None = None,
) -> None:
    if gray:
        color = rcParams["frame.gray.font.color"]
    else:
        color = rcParams[f"frame.{name}.font.color"]
    bold = rcParams[f"frame.{name}.font.bold"]
    size = font_size or rcParams["frame.font.size"]

    set_font(rng, color=color, bold=bold, size=size)


@turn_off_screen_updating
def set_frame_style(
    sf: SheetFrame,
    *,
    autofit: bool = False,
    alignment: str | None = "center",
    banding: bool = False,
    succession: bool = False,
    border: bool = True,
    gray: bool = False,
    font: bool = True,
    fill: bool = True,
    font_size: int | None = None,
) -> None:
    """Set style of SheetFrame.

    Args:
        sf: The SheetFrame object.
        autofit: Whether to autofit the frame.
        alignment: The alignment of the frame.
        border: Whether to draw the border.
        font: Whether to specify the font.
        fill: Whether to fill the frame.
        banding: Whether to draw the banding.
        succession: Whether to hide the succession of the index.
        gray: Whether to set the frame in gray mode.
        font_size: The font size to specify directly.
    """
    cell = sf.cell
    sheet = sf.sheet

    set_style = partial(
        _set_style,
        border=border,
        gray=gray,
        font=font,
        fill=fill,
        font_size=font_size,
    )

    index_level = sf.index_level
    columns_level = sf.columns_level
    length = len(sf)

    if index_level > 0:
        start = cell
        end = cell.offset(columns_level - 1, index_level - 1)
        set_style(start, end, "index.name")

        start = cell.offset(columns_level, 0)
        end = cell.offset(columns_level + length - 1, index_level - 1)
        set_style(start, end, "index")

        if succession:
            rng = sheet.range(start.offset(1, 0), end)
            hide_succession(rng)

            start = cell.offset(columns_level - 1, 0)
            end = cell.offset(columns_level - 1, index_level - 1)
            rng = sheet.range(start, end)
            hide_unique(rng, length)

    width = len(sf.value_columns)

    if columns_level > 1:
        start = cell.offset(0, index_level)
        end = cell.offset(columns_level - 2, index_level + width - 1)
        set_style(start, end, "columns.name")

    start = cell.offset(columns_level - 1, index_level)
    end = cell.offset(columns_level - 1, index_level + width - 1)
    set_style(start, end, "columns")

    start = cell.offset(columns_level, index_level)
    end = cell.offset(columns_level + length - 1, index_level + width - 1)
    set_style(start, end, "values")

    rng = sheet.range(start, end)

    if banding and not gray:
        set_banding(rng)

    rng = sheet.range(cell, end)

    if border:
        ew = 2 if gray else 3
        ec = rcParams["frame.gray.border.color"] if gray else 0
        set_border(rng, edge_weight=ew, inside_weight=0, edge_color=ec)

    if autofit:
        rng.columns.autofit()

    if alignment:
        set_alignment(rng, alignment)


def set_wide_column_style(sf: SheetFrame, gray: bool = False) -> None:
    wide_columns = sf.wide_columns
    edge_color = rcParams["frame.gray.border.color"] if gray else 0

    for wide_column in wide_columns:
        rng = sf.range(wide_column, 0).offset(-1)

        er = 3 if wide_column == wide_columns[-1] else 2
        edge_weight = (1, er - 1, 1, 1) if gray else (2, er, 2, 2)
        set_border(rng, edge_weight, inside_weight=1, edge_color=edge_color)

        _set_style_fill(rng, "wide-columns", gray=gray)
        _set_style_font(rng, "wide-columns", gray=gray)

    for wide_column in wide_columns:
        rng = sf.range(wide_column, 0).offset(-2)

        el = 3 if wide_column == wide_columns[0] else 2
        edge_weight = (el - 1, 2, 2, 1) if gray else (el, 3, 3, 2)
        set_border(rng, edge_weight, inside_weight=0, edge_color=edge_color)

        _set_style_fill(rng, "wide-columns.name", gray=gray)
        _set_style_font(rng, "wide-columns.name", gray=gray)


def set_table_style(
    table: Table,
    even_color: int | str = rgb(240, 250, 255),
    odd_color: int | str = rgb(255, 255, 255),
) -> None:
    book = table.sheet.book.api

    try:
        style = book.TableStyles("xlviews")
    except pywintypes.com_error:
        style = book.TableStyles.Add("xlviews")
        odd_type = TableStyleElementType.xlRowStripe1
        style.TableStyleElements(odd_type).Interior.Color = odd_color
        even_type = TableStyleElementType.xlRowStripe2
        style.TableStyleElements(even_type).Interior.Color = even_color

    table.api.TableStyle = style


def color_palette(n: int) -> list[tuple[int, int, int]]:
    """Return a list of colors of length n."""
    palette = sns.color_palette()
    palette = palette[:n] if n <= len(palette) else sns.husl_palette(n, l=0.5)
    return [tuple(int(c * 255) for c in p) for p in palette]  # type: ignore


MARKER_DICT: dict[str, int] = {
    "o": MarkerStyle.xlMarkerStyleCircle,
    "^": MarkerStyle.xlMarkerStyleTriangle,
    "s": MarkerStyle.xlMarkerStyleSquare,
    "d": MarkerStyle.xlMarkerStyleDiamond,
    "+": MarkerStyle.xlMarkerStylePlus,
    "x": MarkerStyle.xlMarkerStyleX,
    ".": MarkerStyle.xlMarkerStyleDot,
    "-": MarkerStyle.xlMarkerStyleDash,
    "*": MarkerStyle.xlMarkerStyleStar,
}

LINE_DICT: dict[str, int] = {
    "-": LineStyle.xlContinuous,
    "--": LineStyle.xlDash,
    "-.": LineStyle.xlDashDot,
    ".": LineStyle.xlDot,
}


def get_marker_style(marker: int | str | None) -> int:
    if isinstance(marker, int):
        return marker

    if marker is None:
        return MarkerStyle.xlMarkerStyleNone

    return MARKER_DICT[marker]


def get_line_style(line: int | str | None) -> int:
    if isinstance(line, int):
        return line

    if line is None:
        return LineStyle.xlLineStyleNone

    return LINE_DICT[line]


def marker_palette(n: int) -> list[str]:
    """Return a list of markers of length n."""
    return list(itertools.islice(itertools.cycle(MARKER_DICT), n))


def palette(name: str, n: int) -> list[str] | list[tuple[int, int, int]] | list[None]:
    if name == "color":
        return color_palette(n)

    if name == "marker":
        return marker_palette(n)

    return [None] * n


def get_axis_label(axis) -> str | None:  # noqa: ANN001
    if not axis.HasTitle:
        return None

    return axis.AxisTitle.Text


def set_axis_label(
    axis,  # noqa: ANN001
    label: str | tuple[int, int] | Range | None = None,
    name: str | None = None,
    size: float | None = None,
    sheet: Sheet | None = None,
    **kwargs,
) -> None:
    if not label:
        axis.HasTitle = False
        return

    axis.HasTitle = True
    axis_title = axis.AxisTitle
    axis_title.Text = reference(label, sheet)
    size = size or rcParams["chart.axis.title.font.size"]

    set_font_api(axis_title, name, size=size, **kwargs)


def get_ticks(axis) -> tuple[float, float, float, float]:  # noqa: ANN001
    return (
        axis.MinimumScale,
        axis.MaximumScale,
        axis.MajorUnit,
        axis.MinorUnit,
    )


def set_ticks(
    axis,  # noqa: ANN001
    *args,
    min: float | None = None,  # noqa: A002
    max: float | None = None,  # noqa: A002
    major: float | None = None,
    minor: float | None = None,
    gridlines: bool = True,
) -> None:
    args = [*args, None, None, None, None][:4]

    min = min or args[0]  # noqa: A001
    max = max or args[1]  # noqa: A001
    major = major or args[2]
    minor = minor or args[3]

    if min is not None:
        axis.MinimumScale = min

    if max is not None:
        axis.MaximumScale = max

    if major is not None:
        axis.MajorUnit = major

        if gridlines:
            axis.HasMajorGridlines = True
        else:
            axis.HasMajorGridlines = False

    if minor is not None:
        axis.MinorUnit = minor

        if gridlines:
            axis.HasMinorGridlines = True
        else:
            axis.HasMinorGridlines = False

    if min:
        axis.CrossesAt = min


def set_tick_labels(
    axis,  # noqa: ANN001
    name: str | None = None,
    size: float | None = None,
    number_format: str | None = None,
) -> None:
    size = size or rcParams["chart.axis.ticklabels.font.size"]
    set_font_api(axis.TickLabels, name, size=size)

    if number_format:
        axis.TickLabels.NumberFormatLocal = number_format


def get_axis_scale(axis) -> str:  # noqa: ANN001
    if axis.ScaleType == ScaleType.xlScaleLogarithmic:
        return "log"

    if axis.ScaleType == ScaleType.xlScaleLinear:
        return "linear"

    raise NotImplementedError


def set_axis_scale(axis, scale: str) -> None:  # noqa: ANN001
    if scale == "log":
        axis.ScaleType = ScaleType.xlScaleLogarithmic
        return

    if scale == "linear":
        axis.ScaleType = ScaleType.xlScaleLinear
        return

    raise NotImplementedError


def set_dimensions(
    api,  # noqa: ANN001
    left: float | None = None,
    top: float | None = None,
    width: float | None = None,
    height: float | None = None,
) -> None:
    if left is not None:
        api.Left = left

    if top is not None:
        api.Top = top

    if width is not None:
        api.Width = width

    if height is not None:
        api.Height = height


def set_area_format(
    api,  # noqa: ANN001
    border: str | int | tuple[int, int, int] | None = None,
    fill: str | int | tuple[int, int, int] | None = None,
    alpha: float | None = None,
) -> None:
    if border is not None:
        api.Format.Line.Visible = True
        api.Format.Line.ForeColor.RGB = rgb(border)

    if fill is not None:
        api.Format.Fill.Visible = True
        api.Format.Fill.ForeColor.RGB = rgb(fill)

    if alpha is not None:
        api.Format.Line.Transparency = alpha
        api.Format.Fill.Transparency = alpha



import numpy as np
import pytest
from pandas import DataFrame, Series
from xlwings import Sheet

from xlviews.sheetframe import SheetFrame


@pytest.fixture(scope="module")
def df():
    return DataFrame({"a": [1, 2, 3, 4], "b": [5, 6, 7, 8]})


@pytest.fixture(scope="module")
def sf(df: DataFrame, sheet_module: Sheet):
    return SheetFrame(2, 3, data=df, style=False, sheet=sheet_module)


def test_value(sf: SheetFrame):
    v = [[None, "a", "b"], [0, 1, 5], [1, 2, 6], [2, 3, 7], [3, 4, 8]]
    assert sf.cell.expand().options(ndim=2).value == v


def test_init(sf: SheetFrame, sheet_module: Sheet):
    assert sf.row == 2
    assert sf.column == 3
    assert sf.sheet.name == sheet_module.name
    assert sf.index_level == 1
    assert sf.columns_level == 1


def test_set_data_from_sheet(sf: SheetFrame):
    sf.set_data_from_sheet(index_level=1, number_format="0")
    assert sf.has_index is True
    assert sf.index_level == 1


def test_len(sf: SheetFrame):
    assert len(sf) == 4


def test_columns(sf: SheetFrame):
    assert sf.columns == [None, "a", "b"]


def test_value_columns(sf: SheetFrame):
    assert sf.value_columns == ["a", "b"]


def test_index_columns(sf: SheetFrame):
    assert sf.index_columns == [None]


def test_init_index_false(df: DataFrame, sheet: Sheet):
    sf = SheetFrame(2, 3, data=df, index=False, style=False, sheet=sheet)
    assert sf.columns == ["a", "b"]
    assert sf.index_level == 0


def test_row_one(sheet: Sheet):
    df = DataFrame({"a": [1], "b": [2]})
    sf = SheetFrame(2, 2, data=df, style=False, sheet=sheet)
    assert len(sf) == 1
    np.testing.assert_array_equal(sf["a"], [1])


def test_column_one(sheet: Sheet):
    df = DataFrame({"a": [1, 2, 3]})
    sf = SheetFrame(2, 2, data=df, style=False, index=False, sheet=sheet)
    assert len(sf) == 3
    assert sf.columns == ["a"]
    np.testing.assert_array_equal(sf["a"], [1, 2, 3])


def test_contains(sf: SheetFrame):
    assert "a" in sf
    assert "x" not in sf


@pytest.mark.parametrize(
    ("column", "relative", "index"),
    [
        ("a", True, 2),
        ("a", False, 4),
        ("b", True, 3),
        ("b", False, 5),
        (["a", "b"], True, [2, 3]),
        (["a", "b"], False, [4, 5]),
    ],
)
def test_index(sf: SheetFrame, column, relative, index):
    assert sf.index(column, relative=relative) == index


def test_data(sf: SheetFrame, df: DataFrame):
    df_ = sf.data
    np.testing.assert_array_equal(df_.index, df.index)
    np.testing.assert_array_equal(df_.index.names, df.index.names)
    np.testing.assert_array_equal(df_.columns, df.columns)
    np.testing.assert_array_equal(df_.columns.names, df.columns.names)
    np.testing.assert_array_equal(df_, df)
    assert df_.index.name == df.index.name
    assert df_.columns.name == df.columns.name


def test_range_all(sf: SheetFrame):
    assert sf.range_all().get_address() == "$C$2:$E$6"
    assert sf.range().get_address() == "$C$2:$E$6"


@pytest.mark.parametrize(
    ("start", "end", "address"),
    [
        (False, None, "$C$2:$C$6"),
        (-1, None, "$C$2"),
        (0, None, "$C$3"),
        (None, None, "$C$3:$C$6"),
        (10, None, "$C$10"),
        (10, 100, "$C$10:$C$100"),
    ],
)
def test_range_index(sf: SheetFrame, start, end, address):
    assert sf.range_index(start, end).get_address() == address
    assert sf.range("index", start, end).get_address() == address


@pytest.mark.parametrize(
    ("column", "start", "end", "address"),
    [
        ("a", -1, None, "$D$2"),
        ("b", -1, None, "$E$2"),
        ("a", 1, None, "$D$1"),
        ("a", 2, None, "$D$2"),
        ("b", 100, None, "$E$100"),
        ("a", None, None, "$D$3:$D$6"),
        ("a", False, None, "$D$2:$D$6"),
        ("b", False, None, "$E$2:$E$6"),
        ("a", 2, 100, "$D$2:$D$100"),
    ],
)
def test_range_column(sf: SheetFrame, column, start, end, address):
    assert sf.range_column(column, start, end).get_address() == address
    assert sf.range(column, start, end).get_address() == address


def test_repr(sf: SheetFrame):
    assert repr(sf).endswith("!$C$2:$E$6>")


def test_str(sf: SheetFrame):
    assert str(sf).endswith("!$C$2:$E$6>")


def test_rename(sheet: Sheet):
    df = DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
    sf = SheetFrame(2, 2, data=df, style=False, sheet=sheet)
    sf.rename({"a": "A", "b": "B"})
    assert sf.columns == [None, "A", "B"]


def test_drop_duplicates(sheet: Sheet):
    df = DataFrame({"a": [1, 1, 2, 2, 3, 3], "b": [4, 4, 4, 5, 5, 5]})
    sf = SheetFrame(2, 2, data=df, style=False, sheet=sheet)
    sf.drop_duplicates(["a", "b"])
    x = sf["a"]
    assert x[::2].to_list() == [1, 2, 3]
    assert x[1::2].isna().all()
    x = sf["b"]
    assert x[::3].to_list() == [4, 5]
    assert x[1::3].isna().all()
    assert x[2::3].isna().all()


def test_address(sf: SheetFrame):
    assert sf.get_address("a") == ["$D$3", "$D$4", "$D$5", "$D$6"]


def test_address_formula(sf: SheetFrame):
    assert sf.get_address("a", formula=True) == ["=$D$3", "=$D$4", "=$D$5", "=$D$6"]


def test_add_column(sheet: Sheet):
    df = DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
    sf = SheetFrame(2, 2, data=df, style=False, sheet=sheet)
    rng = sf.add_column("c")
    assert rng.get_address() == "$E$3:$E$5"
    assert sf.columns == [None, "a", "b", "c"]


def test_add_column_value(sheet: Sheet):
    df = DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
    sf = SheetFrame(2, 2, data=df, style=False, sheet=sheet)
    sf.add_column("c", [7, 8, 9])
    np.testing.assert_array_equal(sf["c"], [7, 8, 9])


def test_getitem_str(sf: SheetFrame):
    s = sf["a"]
    assert isinstance(s, Series)
    assert s.name == "a"
    np.testing.assert_array_equal(s, [1, 2, 3, 4])


def test_getitem_list(sf: SheetFrame):
    df = sf[["a", "b"]]
    assert isinstance(df, DataFrame)
    assert df.columns.to_list() == ["a", "b"]
    x = [[1, 5], [2, 6], [3, 7], [4, 8]]
    np.testing.assert_array_equal(df, x)


def test_setitem(sheet: Sheet):
    df = DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
    sf = SheetFrame(2, 2, data=df, style=False, sheet=sheet)
    x = [10, 20, 30]
    sf["a"] = x
    np.testing.assert_array_equal(sf["a"], x)


def test_setitem_new_column(sheet: Sheet):
    df = DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
    sf = SheetFrame(2, 2, data=df, style=False, sheet=sheet)
    x = [10, 20, 30]
    sf["c"] = x
    assert sf.columns == [None, "a", "b", "c"]
    np.testing.assert_array_equal(sf["c"], x)


@pytest.mark.parametrize(
    ("a", "b", "sel"),
    [
        (1, None, [True, False, False, False]),
        (3, None, [False, False, True, False]),
        ([2, 4], None, [False, True, False, True]),
        ((2, 4), None, [False, True, True, True]),
        (1, 5, [True, False, False, False]),
        (1, 6, [False, False, False, False]),
        ((1, 3), (6, 8), [False, True, True, False]),
    ],
)
def test_select(sf: SheetFrame, a, b, sel):
    if b is None:
        np.testing.assert_array_equal(sf.select(a=a), sel)
    else:
        np.testing.assert_array_equal(sf.select(a=a, b=b), sel)


def test_groupby(sheet: Sheet):
    df = DataFrame({"a": [1, 1, 1, 2, 2, 1, 1], "b": [1, 2, 3, 4, 5, 6, 7]})
    sf = SheetFrame(2, 2, data=df, style=False, index=False, sheet=sheet)

    g = sf.groupby("a")
    assert len(g) == 2
    assert g[(1,)] == [(3, 5), (8, 9)]
    assert g[(2,)] == [(6, 7)]

    assert len(sf.groupby(["a", "b"])) == 7

    g = sf.groupby("::b")
    assert len(g) == 2
    assert g[(1,)] == [(3, 5), (8, 9)]
    assert g[(2,)] == [(6, 7)]

    assert len(sf.groupby(":b")) == 7


def test_groupby_range(sheet: Sheet):
    df = DataFrame({"a": [1, 1, 1, 2, 2, 1, 1], "b": [3, 3, 4, 4, 3, 3, 4]})
    sf = SheetFrame(2, 2, data=df, style=False, index=False, sheet=sheet)

    g = sf.groupby("a")
    assert sf.range("a", g[(1,)]).get_address() == "$B$3:$B$5,$B$8:$B$9"
    assert sf.range("a", g[(2,)]).get_address() == "$B$6:$B$7"

    g = sf.groupby("b")
    assert sf.range("b", g[(3,)]).get_address() == "$C$3:$C$4,$C$7:$C$8"
    assert sf.range("b", g[(4,)]).get_address() == "$C$5:$C$6,$C$9"

    g = sf.groupby(["a", "b"])
    assert sf.range("a", g[(1, 3)]).get_address() == "$B$3:$B$4,$B$8"
    assert sf.range("a", g[(1, 4)]).get_address() == "$B$5,$B$9"
    assert sf.range("a", g[(2, 3)]).get_address() == "$B$7"
    assert sf.range("a", g[(2, 4)]).get_address() == "$B$6"



"""Modify a SheetFrame."""

from __future__ import annotations

from typing import TYPE_CHECKING

from xlwings.constants import Direction

from xlviews.decorators import turn_off_screen_updating
from xlviews.utils import int_to_column_name, iter_columns

if TYPE_CHECKING:
    from xlwings import Range, Sheet

    from xlviews.sheetframe import SheetFrame


def _move_down(sf: SheetFrame, count: int) -> Range:
    start = sf.row - 1
    end = start + count - 1

    if sf.cell.offset(-1).formula:
        end += 1

    rows = sf.sheet.api.Rows(f"{start}:{end}")
    rows.Insert(Shift=Direction.xlDown)

    return sf.sheet.range(start + 1, sf.column)


def _move_right(sf: SheetFrame, count: int, width: int) -> Range:
    start = sf.column - 1
    end = start + count - 1

    start_name = int_to_column_name(start)
    end_name = int_to_column_name(end)
    columns_name = f"{start_name}:{end_name}"

    columns = sf.sheet.api.Columns(columns_name)
    columns.Insert(Shift=Direction.xlToRight)

    if width:
        columns = sf.sheet.api.Columns(columns_name)
        columns.ColumnWidth = width

    return sf.sheet.range(sf.row, start + 1)


def move(sf: SheetFrame, count: int, direction: str = "down", width: int = 0) -> Range:
    """Insert empty rows/columns to move the SheetFrame to the right or down.

    Args:
        count (int): The number of empty rows/columns to insert.
        direction (str): 'down' or 'right'
        width (int, optional): The width of the columns to insert.

    Returns:
        Range: Original cell.
    """

    match direction:
        case "down":
            return _move_down(sf, count)

        case "right":
            return _move_right(sf, count, width)

    raise ValueError("direction must be 'down' or 'right'")


def delete(sf: SheetFrame, direction: str = "up", *, entire: bool = False) -> None:
    """Delete the SheetFrame.

    Args:
        direction (str): 'up' or 'left'
        entire (bool): Whether to delete the entire row/column.
    """
    rng = sf.range()
    start = rng[0].offset(-1, -1)
    end = rng[-1].offset(1, 1)

    if sf.wide_columns:
        start = start.offset(-1)

    api = sf.sheet.range(start, end).api

    match direction:
        case "up":
            if entire:
                api.EntireRow.Delete()
            else:
                api.Delete(Shift=Direction.xlUp)

        case "left":
            if entire:
                api.EntireColumn.Delete()
            else:
                api.Delete(Shift=Direction.xlToLeft)

        case _:
            raise ValueError("direction must be 'up' or 'left'")


# def get_sheet_cell(sf: SheetFrame, *args) -> tuple[Sheet, Range]:
#     if len(args) == 0:
#         cell = sf.get_adjacent_cell()
#     elif isinstance(args[0], Sheet):
#         cell = common.get_range(*args[1:], sheet=args[0])
#     else:
#         cell = common.get_range(*args)

#     return cell.sheet, cell


# def get_index_level(sf: SheetFrame, columns: list[str]) -> int:
#     index_columns = sf.index_columns
#     level = next(
#         (i for i, c in enumerate(columns) if c not in index_columns), len(columns)
#     )

#     if columns[-1] in index_columns:
#         index_level += 1

#     return index_level


@turn_off_screen_updating
def copy(
    sf: SheetFrame,
    *args,
    columns: list[str] | None = None,
    n=1,
    header_ref=False,
    sort_index=False,
    sel=None,
    rows=None,
    drop_duplicates=False,
    autofit=True,
    sheet: Sheet | None = None,
    **kwargs,
) -> SheetFrame:
    """
    自分の参照コピーを作成する。

    Parameters
    ----------
    *args :
        SheetFrameの第一引数。コピー先の場所を指定する。
    columns : list of str, optional
        コピーするカラム名
    n : int, optional
        行の展開数
    header_ref : bool, optional
        ヘッダー行を参照するか
    sort_index : bool, optional
        インデックスをソートするか
    sel : dict or list of bool, optional
        コピーする行を選択する辞書かファンシーインデックスで指定
    rows: list of int, optional
        コピーする行を行番号で指定. 0-index
    drop_duplicates : bool, optional
        重複するインデックスを削除するか
    autofit : bool, optional
        オートフィットするか

    Returns
    -------
    SheetFrame
    """
    sheet = sheet or sf.sheet
    cell = sheet.range(*args)
    include_sheetname = sf.sheet != cell.sheet
    columns = sf.columns if columns is None else list(iter_columns(sf, columns))
    index_columns = sf.index_columns

    for index_level, column in enumerate(columns):
        if column not in index_columns:
            break
    if columns[-1] in index_columns:
        index_level += 1

    index_data = sf[columns[:index_level]]

    if isinstance(sel, dict):
        sel = sf.select(**sel)
    if sel is not None:
        index_data = index_data[sel]
    if rows:
        index_data = index_data[index_data.index.isin(rows)]
    if drop_duplicates:
        index_data = index_data.drop_duplicates()
    if sort_index:
        index_data = index_data.sort_values(columns[:index_level])

    header = []
    header_cell = {}
    for column in columns:
        if column not in sf.columns:
            header.append(column)
        else:
            header_ = sf.range(column, 0)
            header_cell[column] = header_
            if header_ref:
                header_ = header_.get_address(include_sheetname=include_sheetname)
                header.append("=" + header_)
            else:
                header.append(column)
    values = [header]
    for row in index_data.index:
        row_values = []
        for column in columns:
            if column in header_cell:
                ref = header_cell[column].offset(int(row) + 1)
                ref = ref.get_address(include_sheetname=include_sheetname)
                value = "=" + ref
            else:
                value = None
            row_values.append(value)
        row_values = [row_values] * n
        values.extend(row_values)

    cell.value = values

    self_columns = sf.columns
    number_format = {
        column: sf.get_number_format(column)
        for column in columns
        if column in self_columns
    }

    sf = sf.__class__(cell, index_level=index_level, autofit=False, **kwargs)

    sf.set_number_format(autofit=False, **number_format)

    if autofit:
        sf.range().columns.autofit()

    return sf


# @wait_updating
# def product(self, *args, columns=None, **kwargs):
#     """
#     直積シーﾄフレームを生成する。

#     sf.product(a=[1,2,3], b=[4,5,6])とすると、元のシートフレームを
#     9倍に伸ばして、(1,4), (1,5), ..., (3,6)のデータを追加する.

#     Parameters
#     ----------
#     columns: list
#         積をとるカラム名

#     Returns
#     -------
#     SheetFrame
#     """
#     values = []
#     for value in product(*kwargs.values()):
#         values.append(value)
#     df = pd.DataFrame(values, columns=kwargs.keys())
#     if columns is None:
#         columns = self.columns
#     columns += list(df.columns)
#     length = len(self)
#     sf = self.copy(*args, columns=columns, n=len(df))
#     for column in df:
#         sf[column] = list(df[column]) * length
#     sf.set_style(autofit=True)
#     return sf



from __future__ import annotations

import re
from typing import TYPE_CHECKING

from win32com.client import gencache, selecttlb  # type: ignore

if TYPE_CHECKING:
    from collections.abc import Iterator

    from win32com.client.selecttlb import TypelibSpec


def iter_typelib_specs() -> Iterator[TypelibSpec]:
    pattern = r"Microsoft (Office|Excel|PowerPoint) \S+? Object Library"

    for tlb in selecttlb.EnumTlbs():
        if re.match(pattern, tlb.desc):
            yield tlb


def ensure_module(tlb: TypelibSpec) -> None:
    major = int(tlb.major, 16)
    minor = int(tlb.minor, 16)
    gencache.EnsureModule(tlb.clsid, tlb.lcid, major, minor)  # type: ignore


def ensure_modules() -> None:
    for tlb in iter_typelib_specs():
        ensure_module(tlb)



import pytest
from pandas import DataFrame
from xlwings import Sheet

from xlviews.sheetframe import SheetFrame


@pytest.fixture(scope="module")
def df():
    df = DataFrame(
        {
            "x": [1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2],
            "y": [3, 3, 3, 3, 3, 4, 4, 4, 4, 3, 3, 3, 4, 4],
            "a": [5, 4, 3, 2, 1, 4, 3, 2, 1, 3, 2, 1, 2, 1],
            "b": [1, 2, 3, 4, 5, 1, 2, 3, 4, 1, 2, 3, 1, 2],
        },
    )
    return df.set_index(["x", "y"])


@pytest.fixture(scope="module")
def sf(df: DataFrame, sheet_module: Sheet):
    return SheetFrame(3, 2, data=df, style=False, sheet=sheet_module)



import pytest
from pandas import DataFrame
from xlwings import Sheet

from xlviews.sheetframe import SheetFrame


@pytest.fixture
def sfs(sheet: Sheet):
    df0 = DataFrame({"a": [1, 1], "b": [1, 1]})
    df0 = df0.set_index("a")
    df1 = DataFrame({"c": [2, 2], "d": [2, 2]})
    df1 = df1.set_index("c")
    df2 = DataFrame({"e": [3, 3], "f": [3, 3]})
    df2 = df2.set_index("e")

    sf0 = SheetFrame(2, 3, data=df0, style=False, sheet=sheet)
    sf1 = SheetFrame(7, 3, data=df1, style=False, sheet=sheet)
    sf2 = SheetFrame(2, 7, data=df2, style=False, sheet=sheet)

    return [sf0, sf1, sf2]


def test_sfs_value(sfs: list[SheetFrame]):
    v = sfs[0].sheet.range("$C$2:$I$10").value
    assert isinstance(v, list)
    assert len(v) == 9
    assert v[0] == ["a", "b", None, None, "e", "f", None]
    assert v[1] == [1, 1, None, None, 3, 3, None]
    assert v[2] == [1, 1, None, None, 3, 3, None]
    assert v[3] == [None] * 7
    assert v[4] == [None] * 7
    assert v[5] == ["c", "d", None, None, None, None, None]
    assert v[6] == [2, 2, None, None, None, None, None]
    assert v[7] == [2, 2, None, None, None, None, None]
    assert v[8] == [None] * 7


def test_delete_up(sfs: list[SheetFrame]):
    sfs[0].delete("up")

    v = sfs[0].sheet.range("$C$2:$I$10").value
    assert isinstance(v, list)
    assert len(v) == 9
    assert v[0] == ["c", "d", None, None, "e", "f", None]
    assert v[1] == [2, 2, None, None, 3, 3, None]
    assert v[2] == [2, 2, None, None, 3, 3, None]
    assert v[3] == [None] * 7
    assert v[4] == [None] * 7
    assert v[5] == [None] * 7
    assert v[6] == [None] * 7
    assert v[7] == [None] * 7
    assert v[8] == [None] * 7


def test_delete_left(sfs: list[SheetFrame]):
    sfs[0].delete("left")

    v = sfs[0].sheet.range("$C$2:$I$10").value
    assert isinstance(v, list)
    assert len(v) == 9
    assert v[0] == ["e", "f", None, None, None, None, None]
    assert v[1] == [3, 3, None, None, None, None, None]
    assert v[2] == [3, 3, None, None, None, None, None]
    assert v[3] == [None] * 7
    assert v[4] == [None] * 7
    assert v[5] == ["c", "d", None, None, None, None, None]
    assert v[6] == [2, 2, None, None, None, None, None]
    assert v[7] == [2, 2, None, None, None, None, None]
    assert v[8] == [None] * 7


def test_delete_up_entire(sfs: list[SheetFrame]):
    sfs[0].delete("up", entire=True)

    v = sfs[0].sheet.range("$C$2:$I$10").value
    assert isinstance(v, list)
    assert len(v) == 9
    assert v[0] == ["c", "d", None, None, None, None, None]
    assert v[1] == [2, 2, None, None, None, None, None]
    assert v[2] == [2, 2, None, None, None, None, None]
    assert v[3] == [None] * 7
    assert v[4] == [None] * 7
    assert v[5] == [None] * 7
    assert v[6] == [None] * 7
    assert v[7] == [None] * 7
    assert v[8] == [None] * 7


def test_delete_left_entire(sfs: list[SheetFrame]):
    sfs[0].delete("left", entire=True)

    v = sfs[0].sheet.range("$C$2:$I$10").value
    assert isinstance(v, list)
    assert len(v) == 9
    assert v[0] == ["e", "f", None, None, None, None, None]
    assert v[1] == [3, 3, None, None, None, None, None]
    assert v[2] == [3, 3, None, None, None, None, None]
    assert v[3] == [None] * 7
    assert v[4] == [None] * 7
    assert v[5] == [None] * 7
    assert v[6] == [None] * 7
    assert v[7] == [None] * 7
    assert v[8] == [None] * 7


def test_delete_error(sfs: list[SheetFrame]):
    with pytest.raises(ValueError, match="direction must be 'up' or 'left'"):
        sfs[0].delete("right")



import numpy as np
import pytest
from pandas import DataFrame, Series
from xlwings import Sheet

from xlviews.sheetframe import SheetFrame
from xlviews.table import Table


@pytest.fixture(scope="module")
def df():
    df = DataFrame({"a": [1, 2, 3, 4], "b": [5, 6, 7, 8]}, index=["x", "x", "y", "y"])
    df.index.name = "name"
    return df


@pytest.fixture(scope="module")
def sf(df: DataFrame, sheet_module: Sheet):
    return SheetFrame(2, 3, data=df, style=False, sheet=sheet_module)


def test_value(sf: SheetFrame):
    v = [["name", "a", "b"], ["x", 1, 5], ["x", 2, 6], ["y", 3, 7], ["y", 4, 8]]
    assert sf.cell.expand().options(ndim=2).value == v


def test_init(sf: SheetFrame, sheet_module: Sheet):
    assert sf.row == 2
    assert sf.column == 3
    assert sf.sheet.name == sheet_module.name
    assert sf.index_level == 1
    assert sf.columns_level == 1


def test_set_data_from_sheet(sf: SheetFrame):
    sf.set_data_from_sheet(index_level=0)
    assert sf.has_index is False
    assert sf.value_columns == ["name", "a", "b"]
    sf.set_data_from_sheet(index_level=1)
    assert sf.has_index is True
    assert sf.value_columns == ["a", "b"]


def test_columns(sf: SheetFrame):
    assert sf.columns == ["name", "a", "b"]


def test_value_columns(sf: SheetFrame):
    assert sf.value_columns == ["a", "b"]


def test_index_columns(sf: SheetFrame):
    assert sf.index_columns == ["name"]


def test_contains(sf: SheetFrame):
    assert "name" in sf
    assert "a" in sf
    assert "x" not in sf


@pytest.mark.parametrize(
    ("column", "relative", "index"),
    [
        ("name", True, 1),
        ("name", False, 3),
        ("a", True, 2),
        ("a", False, 4),
        ("b", True, 3),
        ("b", False, 5),
        (["name", "b"], True, [1, 3]),
        (["name", "b"], False, [3, 5]),
    ],
)
def test_index(sf: SheetFrame, column, relative, index):
    assert sf.index(column, relative=relative) == index


def test_data(sf: SheetFrame, df: DataFrame):
    df_ = sf.data
    np.testing.assert_array_equal(df_.index, df.index)
    np.testing.assert_array_equal(df_.index.names, df.index.names)
    np.testing.assert_array_equal(df_.columns, df.columns)
    np.testing.assert_array_equal(df_.columns.names, df.columns.names)
    np.testing.assert_array_equal(df_, df)
    assert df_.index.name == df.index.name
    assert df_.columns.name == df.columns.name


@pytest.fixture
def table(sf: SheetFrame):
    yield sf.as_table()
    sf.unlist()


@pytest.mark.parametrize("value", ["x", "y"])
def test_table(table: Table, value):
    table.auto_filter("name", value)
    header = table.const_header.value
    assert isinstance(header, list)
    assert header[0] == value


@pytest.mark.parametrize(
    ("name", "value"),
    [
        ("x", [[1, 5], [2, 6]]),
        ("y", [[3, 7], [4, 8]]),
    ],
)
def test_visible_data(sf: SheetFrame, table: Table, name, value):
    table.auto_filter("name", name)
    df = sf.visible_data
    assert df.index.to_list() == [name, name]
    np.testing.assert_array_equal(df, value)


def test_range_all(sf: SheetFrame):
    assert sf.range_all().get_address() == "$C$2:$E$6"
    assert sf.range().get_address() == "$C$2:$E$6"


@pytest.mark.parametrize(
    ("start", "end", "address"),
    [
        (False, None, "$C$2:$C$6"),
        (-1, None, "$C$2"),
        (0, None, "$C$3"),
        (None, None, "$C$3:$C$6"),
        (10, None, "$C$10"),
        (10, 100, "$C$10:$C$100"),
    ],
)
def test_range_index(sf: SheetFrame, start, end, address):
    assert sf.range_index(start, end).get_address() == address
    assert sf.range("index", start, end).get_address() == address


@pytest.mark.parametrize(
    ("column", "start", "end", "address"),
    [
        ("a", -1, None, "$D$2"),
        ("b", -1, None, "$E$2"),
        ("name", -1, None, "$C$2"),
        ("a", 1, None, "$D$1"),
        ("a", 2, None, "$D$2"),
        ("b", 100, None, "$E$100"),
        ("a", None, None, "$D$3:$D$6"),
        ("a", False, None, "$D$2:$D$6"),
        ("b", False, None, "$E$2:$E$6"),
        ("name", False, None, "$C$2:$C$6"),
        ("a", 2, 100, "$D$2:$D$100"),
    ],
)
def test_range_column(sf: SheetFrame, column, start, end, address):
    assert sf.range_column(column, start, end).get_address() == address
    assert sf.range(column, start, end).get_address() == address


@pytest.mark.parametrize(
    ("column", "address"),
    [("a", "$D$3"), ("b", "$E$3"), ("name", "$C$3")],
)
def test_first_range(sf: SheetFrame, column, address):
    assert sf.first_range(column).get_address() == address


def test_getitem_str(sf: SheetFrame):
    s = sf["a"]
    assert isinstance(s, Series)
    assert s.name == "a"
    np.testing.assert_array_equal(s, [1, 2, 3, 4])


def test_getitem_list(sf: SheetFrame):
    df = sf[["a", "b"]]
    assert isinstance(df, DataFrame)
    assert df.columns.to_list() == ["a", "b"]
    x = [[1, 5], [2, 6], [3, 7], [4, 8]]
    np.testing.assert_array_equal(df, x)


@pytest.mark.parametrize(
    ("name", "a", "sel"),
    [
        ("x", None, [True, True, False, False]),
        ("y", None, [False, False, True, True]),
        ("x", (2, 3), [False, True, False, False]),
        ("x", [1, 4], [True, False, False, False]),
        (["x", "y"], [1, 4], [True, False, False, True]),
    ],
)
def test_select(sf: SheetFrame, name, a, sel):
    if a is None:
        np.testing.assert_array_equal(sf.select(name=name), sel)
    else:
        np.testing.assert_array_equal(sf.select(name=name, a=a), sel)


def test_groupby(sf: SheetFrame):
    g = sf.groupby("name")
    assert len(g) == 2
    assert g[("x",)] == [(3, 4)]
    assert g[("y",)] == [(5, 6)]

    assert len(sf.groupby(["name", "a"])) == 4


@pytest.fixture(scope="module")
def sf2(sheet_module: Sheet):
    a = ["c"] * 10
    b = ["s"] * 5 + ["t"] * 5
    c = ([100] * 2 + [200] * 3) * 2
    x = list(range(10))
    y = list(range(10, 20))
    df = DataFrame({"a": a, "b": b, "c": c, "x": x, "y": y})
    df = df.set_index(["a", "b", "c"])
    return SheetFrame(102, 2, data=df, index=True, style=False, sheet=sheet_module)


@pytest.mark.parametrize(
    ("kwargs", "r"),
    [({}, range(103, 113)), ({"c": 100}, [103, 104, 108, 109])],
)
def test_ranges(sf2: SheetFrame, kwargs, r):
    for rng, i in zip(sf2.ranges(**kwargs), r, strict=True):
        assert rng.get_address() == f"$E${i}:$F${i}"


def test_ranges_sel(sf2: SheetFrame):
    sel = sf2.select(c=200)
    it = sf2.ranges(sel, b="t")

    for rng, i in zip(it, [110, 111, 112], strict=True):
        assert rng.get_address() == f"$E${i}:$F${i}"



from __future__ import annotations

from xlwings import Range, Sheet

from xlviews.axes import Axes
from xlviews.series import Series
from xlviews.sheetframe import SheetFrame


def _plot(
    ax: Axes,
    data: SheetFrame,
    x: str | tuple,
    y: str | tuple | None = None,
    *,
    label: str | tuple[int, int] | Range = "",
    chart_type: int | None = None,
) -> Series:
    print(data.range(x))
    # if label is None:
    #     label = ax.sheet.range("A1")
    # s = ax.add_series(data[x], data[y], label=label)
    # s.set(marker="o", line="-", edge_weight=1.25, line_weight=1.25)
    # s.delete()



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from pandas import DataFrame
from xlwings import Range
from xlwings.constants import Direction

from xlviews.config import rcParams
from xlviews.decorators import turn_off_screen_updating
from xlviews.formula import AGG_FUNCS, aggregate
from xlviews.grouper import Grouper
from xlviews.range import RangeCollection, multirange
from xlviews.sheetframe import SheetFrame
from xlviews.style import set_font
from xlviews.utils import iter_columns

if TYPE_CHECKING:
    from collections.abc import Iterator

    from numpy.typing import NDArray


class StatsGrouper(Grouper):
    def ranges(self, column: str) -> Iterator[Range | RangeCollection | None]:
        if column in self.by:
            yield from super().first_ranges(column)

        elif column in self.sf.index_columns:
            yield from [None] * len(self.grouped)

        else:
            yield from super().ranges(column)

    def iter_formulas(
        self,
        column: str,
        funcs: list[str] | dict[str, str],
        wrap: str | None = None,
        default: str = "median",
    ) -> Iterator[str]:
        for ranges in self.ranges(column):
            if isinstance(funcs, dict):
                funcs = [funcs.get(column, default)]

            for func in funcs:
                yield get_formula(func, ranges, wrap)

    def get_index(self, funcs: list[str]) -> list[str]:
        return funcs * len(self.grouped)

    def get_columns(
        self,
        funcs: list[str] | dict[str, str],
        func_column_name: str = "func",
    ) -> list[str]:
        columns = self.sf.columns

        if isinstance(funcs, list):
            columns = [func_column_name, *columns]

        return columns

    def get_values(
        self,
        funcs: list[str] | dict[str, str],
        wrap: str | dict[str, str] | None = None,
        default: str = "median",
    ) -> NDArray[np.str_]:
        values = [self.get_index(funcs)] if isinstance(funcs, list) else []

        for column in self.sf.columns:
            wrap_ = wrap.get(column) if isinstance(wrap, dict) else wrap
            it = self.iter_formulas(column, funcs, wrap_, default)
            values.append(list(it))

        return np.array(values).T

    def get_frame(
        self,
        funcs: list[str] | dict[str, str],
        wrap: str | dict[str, str] | None = None,
        default: str = "median",
        func_column_name: str = "func",
    ) -> DataFrame:
        values = self.get_values(funcs, wrap, default)
        columns = self.get_columns(funcs, func_column_name)
        df = DataFrame(values, columns=columns)
        return df.set_index(columns[: -len(self.sf.value_columns)])


def get_formula(
    func: str | Range,
    ranges: Range | RangeCollection | None,
    wrap: str | None = None,
) -> str:
    if not ranges:
        return ""

    if isinstance(ranges, Range):
        return "=" + ranges.get_address()

    formula = aggregate(func, ranges)

    if wrap:
        formula = wrap.format(formula)

    return f"={formula}"


class StatsFrame(SheetFrame):
    parent: SheetFrame

    @turn_off_screen_updating
    def __init__(
        self,
        parent: SheetFrame,
        funcs: str | list[str] | dict[str, str] | None = None,
        *,
        by: str | list[str] | None = None,
        table: bool = True,
        wrap: str | dict[str, str] | None = None,
        na: str | list[str] | bool = False,
        null: str | list[str] | bool = False,
        default: str = "median",
        func_column_name: str = "func",
        succession: bool = False,
        auto_filter: bool = True,
        **kwargs,
    ) -> None:
        """Create a StatsFrame.

        Args:
            parent (SheetFrame): The sheetframe to be aggregated.
            funcs (str, list of str, dict, optional): The aggregation
                functions to be used. The following functions are supported:
                    'count', 'sum', 'min', 'max', 'mean', 'median', 'std', 'soa%'
                None to use the default functions.
            by (str, list of str, optional): The column names to be grouped by.
            autofilter (bool): If True, the displayed functions are limited to
                the default ones.
            table (bool): If True, the frame is displayed in table format.
            wrap (str or dict, optional): A string to wrap the aggregation
                functions. {} is replaced with the aggregation functions.
            na (bool): If True, self.wrap = 'IFERROR({},NA())' is used.
            null (bool): If True, self.wrap = 'IFERROR({},"")' is used.
            succession (bool, optional): If True, the continuous index is hidden.
            **kwargs: Passed to SheetFrame.__init__.
        """
        funcs = get_func(funcs)

        # Store the position of the parent SheetFrame before moving down.
        row = parent.row
        column = parent.column
        if isinstance(funcs, list):
            column -= 1

        by = list(iter_columns(parent, by)) if by else []
        offset = get_length(parent, by, funcs) + 2

        move_down(parent, offset)

        gr = StatsGrouper(parent, by)
        wrap = get_wrap(wrap, na=na, null=null)
        df = gr.get_frame(funcs, wrap, default, func_column_name)

        super().__init__(
            row,
            column,
            data=df,
            index=parent.has_index,
            autofit=False,
            style=False,
            sheet=parent.sheet,
            **kwargs,
        )
        self.parent = parent

        if table:
            self.as_table(autofit=False, const_header=True)

        self.set_style(autofit=False, succession=succession)

        if isinstance(funcs, list):
            self.set_value_style(func_column_name)

        if table:
            self.set_alignment("left")

        if self.table and auto_filter and isinstance(funcs, list) and len(funcs) > 1:
            func = "median" if "median" in funcs else funcs[0]
            self.table.auto_filter(func_column_name, func)

    def set_value_style(self, func_column_name: str) -> None:
        start = self.column + self.index_level
        end = self.column + len(self.columns)
        func_index = self.index(func_column_name)

        value_columns = ["median", "min", "mean", "max", "std", "sum"]

        grouped = self.groupby(func_column_name)
        columns = [func_index, *range(start, end)]
        get_fmt = self.parent.get_number_format
        formats = [get_fmt(column) for column in self.value_columns]
        formats = [None, *formats]

        for key, rows in grouped.items():
            func = key[0]
            for column, fmt in zip(columns, formats, strict=False):
                cell = multirange(self.sheet, rows, column)

                if func in value_columns:
                    cell.number_format = fmt

                color = rcParams.get(f"stats.{func}.color")
                italic = rcParams.get(f"stats.{func}.italic")
                set_font(cell, color=color, italic=italic)

                if func == "soa" and column != func_index:
                    cell.number_format = "0.0%"

        set_font(self.range(func_column_name), italic=True)


def get_wrap(
    wrap: str | dict[str, str] | None = None,
    *,
    na: str | list[str] | bool = False,
    null: str | list[str] | bool = False,
) -> str | dict[str, str] | None:
    if wrap:
        return wrap

    if na is True:
        return "IFERROR({},NA())"

    if null is True:
        return 'IFERROR({},"")'

    wrap = {}

    if na:
        nas = [na] if isinstance(na, str) else na
        for na in nas:
            wrap[na] = "IFERROR({},NA())"

    if null:
        nulls = [null] if isinstance(null, str) else null
        for null in nulls:
            wrap[null] = 'IFERROR({},"")'

    return wrap or None


def get_func(
    func: str | list[str] | dict[str, str] | None,
) -> list[str] | dict[str, str]:
    if func is None:
        func = list(AGG_FUNCS.keys())
        func.remove("sum")
        func.remove("std")
        return func

    return [func] if isinstance(func, str) else func


def get_length(sf: SheetFrame, by: list[str], funcs: list | dict) -> int:
    n = 1 if isinstance(funcs, dict) else len(funcs)

    if not by:
        return n

    return len(sf[by].drop_duplicates()) * n


def has_header(sf: SheetFrame) -> bool:
    start = sf.cell.offset(-1)
    end = start.offset(0, len(sf.columns))
    value = sf.sheet.range(start, end).options(ndim=1).value

    if not isinstance(value, list):
        raise NotImplementedError

    return any(value)


def move_down(sf: SheetFrame, length: int) -> int:
    start = sf.row - 1
    end = sf.row + length - 2

    if has_header(sf):
        end += 1

    rows = sf.sheet.api.Rows(f"{start}:{end}")
    rows.Insert(Shift=Direction.xlDown)
    return end - start + 1



import pytest


def test_sns_color_palette():
    import seaborn as sns

    cp = sns.color_palette()
    assert len(cp) == 10


def test_color_palette():
    from xlviews.style import color_palette

    cp = color_palette(3)
    assert cp == [(31, 119, 180), (255, 127, 14), (44, 160, 44)]


def test_color_palette_large():
    from xlviews.style import color_palette

    cp = color_palette(20)
    assert cp[0] == (224, 37, 89)
    assert cp[-1] == (216, 40, 137)


def test_marker_palette():
    from xlviews.style import marker_palette

    mp = marker_palette(10)
    assert mp[0] == "o"
    assert mp[-2] == "*"
    assert mp[-1] == "o"


@pytest.mark.parametrize(
    ("name", "p"),
    [
        ("color", [(31, 119, 180), (255, 127, 14)]),
        ("marker", ["o", "^"]),
        ("none", [None, None]),
    ],
)
def test_palette_color(name: str, p):
    from xlviews.style import palette

    assert palette(name, 2) == p


def test_marker_style_int():
    from xlviews.style import get_marker_style

    assert get_marker_style(1) == 1


def test_line_style_int():
    from xlviews.style import get_line_style

    assert get_line_style(1) == 1



import numpy as np
import pytest

from xlviews.sheetframe import SheetFrame
from xlviews.statsframe import StatsFrame


@pytest.fixture(scope="module")
def sf(sf_parent: SheetFrame):
    funcs = ["count", "max", "median", "soa"]
    return StatsFrame(sf_parent, funcs, by=":y", table=True)


def test_len(sf: StatsFrame):
    assert len(sf) == 16


def test_columns(sf: StatsFrame):
    assert sf.columns == ["func", "x", "y", "z", "a", "b", "c"]


def test_index_columns(sf: StatsFrame):
    assert sf.index_columns == ["func", "x", "y", "z"]


def test_value_columns(sf: StatsFrame):
    assert sf.value_columns == ["a", "b", "c"]


@pytest.mark.parametrize(
    ("func", "n"),
    [
        ("median", 4),
        (["soa"], 4),
        (["count", "max", "median"], 12),
        (["count", "max", "median", "soa"], 16),
    ],
)
def test_value_len(sf: StatsFrame, func, n):
    assert sf.table
    sf.table.auto_filter("func", func)
    df = sf.visible_data
    assert len(df) == n


@pytest.mark.parametrize(
    ("func", "column", "value"),
    [
        ("count", "a", [7, 3, 4, 4]),
        ("count", "b", [8, 4, 4, 4]),
        ("count", "c", [7, 3, 3, 4]),
        ("max", "a", [18, 7, 11, 15]),
        ("max", "b", [27, 7, 9, 15]),
        ("max", "c", [24, 34, 36, 10]),
        ("median", "a", [3, 6, 9.5, 13.5]),
        ("median", "b", [10.5, 5.5, 5.5, 10.5]),
        ("median", "c", [18, 30, 2, 7]),
    ],
)
def test_value_float(sf: StatsFrame, func, column, value):
    assert sf.table
    sf.table.auto_filter("func", func)
    df = sf.visible_data
    np.testing.assert_allclose(df[column], value)


@pytest.mark.parametrize(
    ("column", "value"),
    [
        ("a", [[0, 1, 2, 3, 16, 17, 18], [5, 6, 7]]),
        ("b", [[0, 1, 2, 3, 18, 21, 24, 27], [4, 5, 6, 7]]),
        ("c", [[20, 22, 24, 12, 14, 16, 18], [28, 30, 34]]),
    ],
)
def test_value_soa(sf: StatsFrame, column, value):
    assert sf.table
    sf.table.auto_filter("func", "soa")
    df = sf.visible_data
    soa = [np.std(x) / np.median(x) for x in value]
    np.testing.assert_allclose(df[column].iloc[:2], soa)



"""
マトリックス上にオブジェクトを配置するために、
FacetGrid, PairGridクラスを定義する。
Seabornの機能をチャート等にも適用できるようにする。
"""

import pandas as pd

# from xlviews.decorators import wait_updating
# from xlviews.chart import chart_position
from xlviews.axes import chart_position

# from spin.pandas import select
from xlviews.config import rcParams


class GridBase:
    def __init__(
        self,
        data,
        sheet=None,
        cell=None,
        left=None,
        top=None,
        width=None,
        height=None,
        row=None,
        column=None,
        space=0,
        callback=None,
    ):
        """
        GridBaseを生成する。

        Parameters
        ----------
        data : SheetFrame or DataFrame
            グリッド上に配置する元データ
        left, top : int, optional
            グリッドの開始位置
        width, height : int, optional
            オブジェクトのサイズ
        row, column : int, optional
            グリッドの開始位置をセルで指定する。
            この指定を行うと、シートフレームをグリッド上に配置する。
            mapで関数を呼び出すとき、func(sheet, row, column, ...)
        space : int or list, optional
            グリッド間の間隔
        """
        if cell is not None:
            sheet, row, column = cell.sheet, cell.row, cell.column
        if isinstance(space, int):
            space = space, space
        self.xspace, self.yspace = space
        self.data = data
        self.sheet = sheet if sheet else data.sheet

        if row and column:
            self.is_cellwise = True
            self.top, self.left = row, column
        else:
            self.is_cellwise = False
            self.left, self.top = chart_position(self.sheet, left, top)
            self.width = width or int(rcParams["chart.width"])
            self.height = height or int(rcParams["chart.height"])

        self.callback = callback
        self.elements = []
        self.element_dict = {}
        self.title = None

    def get_data(self):
        if hasattr(self.data, "columns_names"):
            if self.data.columns_names is None:
                data = self.data
            else:
                data = pd.DataFrame(
                    self.data.value_columns,
                    columns=self.data.columns_names,
                )
        else:
            data = self.data  # DataFrame or DataSet

        return data

    def __getattr__(self, item):
        def func(*args, **kwargs):
            return self.map(item, *args, **kwargs)

        return func

    # def columns_list(self, columns):
    #     if hasattr(self.data, 'columns_list'):
    #         return self.data.columns_list(x) if x else None
    #     else:
    #         return [x] if isinstance(x, str) else x


class FacetGrid(GridBase):
    def __init__(
        self,
        data,
        x=None,
        y=None,
        sheet=None,
        left=None,
        top=None,
        width=None,
        height=None,
        cell=None,
        row=None,
        column=None,
        space=0,
        callback=None,
    ):
        """
        FacetGridを生成する。

        Parameters
        ----------
        data : SheetFrame or DataFrame
            グリッド上に配置する元データ
        sheet
        left, top : int, optional
            グリッドの開始位置
        width, height : int, optional
            オブジェクトのサイズ
        row, column : int, optional
            グリッドの開始位置をセルで指定する。
            この指定を行うと、シートフレームをグリッド上に配置する。
            mapで関数を呼び出すとき、func(sheet, row, column, ...)
        x, y : str, optional
            列/行方向のカラム名
        space : int or list, optional
            グリッド間の間隔
        """
        super().__init__(
            data,
            sheet=sheet,
            cell=cell,
            left=left,
            top=top,
            width=width,
            height=height,
            row=row,
            column=column,
            space=space,
            callback=callback,
        )
        # if hasattr(data, 'columns_list'):
        #     self.x = data.columns_list(x) if x else None
        #     self.y = data.columns_list(y) if y else None
        # else:
        #     self.x = [x] if isinstance(x, str) else x
        #     self.y = [y] if isinstance(y, str) else y
        self.x = data.columns_list(x) if x else None
        self.y = data.columns_list(y) if y else None

        self.title = (self.x or []) + (self.y or [])

    def map(self, plot, *args, sel=None, by=None, title="auto", **kwargs):
        """
        FacetGridの各データに関数funcを適用する。
        funcの第一引数にはデータが渡される。
        """
        data = self.get_data()

        if self.y:
            row_df = data[self.y].drop_duplicates()
        else:
            row_df = pd.DataFrame([[0]], columns=["__dummy__"])
        if self.x:
            col_df = data[self.x].drop_duplicates()
        else:
            col_df = pd.DataFrame([[0]], columns=["__dummy__"])

        sel = sel if sel else {}
        if by is None:
            by = []
        elif not isinstance(by, list):
            by = [by]

        if isinstance(plot, str):
            plot = getattr(self.data, plot)
            data_kw = {}
        else:
            data_kw = {"data": self.data}

        if title == "auto":
            title = self.title

        self.elements = []
        self.element_dict = {}
        top = self.top
        for key_y, row_series in row_df.iterrows():
            if "__dummy__" in row_series:
                del row_series["__dummy__"]
            row_dict = dict(row_series)
            left = self.left
            row_elements = []
            element = None
            for key_x, col_series in col_df.iterrows():
                if "__dummy__" in col_series:
                    del col_series["__dummy__"]
                col_dict = dict(col_series)
                sel_ = dict(**sel, **col_dict, **row_dict)
                if self.is_cellwise:
                    element = plot(
                        self.sheet,
                        top,
                        left,
                        sel=sel_,
                        title=title,
                        **kwargs,
                        **data_kw,
                    )
                else:
                    element = plot(
                        *args,
                        sel=sel_,
                        by=by,
                        left=left,
                        top=top,
                        width=self.width,
                        height=self.height,
                        title=title,
                        sheet=self.sheet,
                        **kwargs,
                        **data_kw,
                    )
                    if element is None:
                        continue
                if self.is_cellwise:
                    left += element.width + 1
                else:
                    left += self.width
                if callable(self.callback):
                    self.callback(element, sel_)
                row_elements.append(element)
                key = []
                if self.x:
                    key += [col_dict[x_] for x_ in self.x]
                if self.y:
                    key += [row_dict[y_] for y_ in self.y]
                self.element_dict[tuple(key)] = element
            if self.is_cellwise:
                top += element.height + 1
            else:
                top += self.height
            self.elements.append(row_elements)
        return self.elements


class PairGrid(GridBase):
    def __init__(
        self,
        data,
        x_vars=None,
        y_vars=None,
        sheet=None,
        left=None,
        top=None,
        width=None,
        height=None,
        cell=None,
        row=None,
        column=None,
        space=0,
        callback=None,
    ):
        """
        PairGridを生成する。

        Parameters
        ----------
        data : SheetFrame or DataFrame
            グリッド上に配置する元データ
        left, top : int
            グリッドの開始位置
        width, height : int, optional
            オブジェクトのサイズ
        row, column : int, optional
            グリッドの開始位置をセルで指定する。
            この指定を行うと、シートフレームをグリッド上に配置する。
            mapで関数を呼び出すとき、func(sheet, row, column, ...)
        x_vars, y_vars : str, optional
            x, y のカラム名
        space : int or list, optional
            グリッド間の間隔
        """
        super().__init__(
            data,
            sheet=sheet,
            cell=cell,
            left=left,
            top=top,
            width=width,
            height=height,
            row=row,
            column=column,
            space=space,
            callback=callback,
        )
        if hasattr(data, "columns_list"):
            self.x_vars = data.columns_list(x_vars) if x_vars else None
            self.y_vars = data.columns_list(y_vars) if y_vars else None
        else:
            self.x_vars = [x_vars] if isinstance(x_vars, str) else x_vars
            self.y_vars = [y_vars] if isinstance(y_vars, str) else y_vars

    def map(
        self,
        plot,
        *args,
        sel=None,
        by=None,
        title="auto",
        transpose=False,
        **kwargs,
    ):
        """
        PairGridの各データに関数funcを適用する。
        funcの第一引数にはデータが渡される。
        """
        data = self.get_data()

        if self.y:
            row_df = data[self.y].drop_duplicates()
        else:
            row_df = pd.DataFrame([[0]], columns=["__dummy__"])
        if self.x:
            col_df = data[self.x].drop_duplicates()
        else:
            col_df = pd.DataFrame([[0]], columns=["__dummy__"])

        sel = sel if sel else {}
        if by is None:
            by = []
        elif not isinstance(by, list):
            by = [by]

        if isinstance(plot, str):
            plot = getattr(self.data, plot)
            data_kw = {}
        else:
            data_kw = {"data": self.data}

        if title == "auto":
            title = self.title

        self.elements = []
        top = self.top
        for key_y, row_series in row_df.iterrows():
            if "__dummy__" in row_series:
                del row_series["__dummy__"]
            row_dict = dict(row_series)
            left = self.left
            row_elements = []
            element = None
            for key_x, col_series in col_df.iterrows():
                if "__dummy__" in col_series:
                    del col_series["__dummy__"]
                col_dict = dict(col_series)
                sel_ = dict(**sel, **col_dict, **row_dict)
                if self.is_cellwise:
                    element = plot(
                        self.sheet,
                        top,
                        left,
                        sel=sel_,
                        title=title,
                        **kwargs,
                        **data_kw,
                    )
                else:
                    element = plot(
                        *args,
                        sel=sel_,
                        by=by,
                        left=left,
                        top=top,
                        width=self.width,
                        height=self.height,
                        title=title,
                        **kwargs,
                        **data_kw,
                    )
                    if element is None:
                        continue
                if self.is_cellwise:
                    left += element.width + 1
                else:
                    left += self.width
                if callable(self.callback):
                    self.callback(element, sel_)
                row_elements.append(element)
            if self.is_cellwise:
                top += element.height + 1
            else:
                top += self.height
            self.elements.append(row_elements)
        return self.elements


def main():
    # import xlwings as xw
    import xlviews as xv

    sf = xv.SheetFrame(2, 2, index_level=3, style=False)
    print(sf)


if __name__ == "__main__":
    main()



if __name__ == "__main__":
    import numpy as np
    import xlwings as xw
    from pandas import DataFrame, MultiIndex
    from xlwings.constants import ChartType

    from xlviews.axes import Axes
    from xlviews.sheetframe import SheetFrame

    for app in xw.apps:
        app.quit()

    book = xw.Book()
    sheet_module = book.sheets.add()

    a = ["a"] * 8 + ["b"] * 8
    b = (["c"] * 4 + ["d"] * 4) * 2
    c = np.repeat(range(1, 9), 2)
    d = ["x", "y"] * 8
    df = DataFrame(np.arange(16 * 6).reshape(16, 6).T)
    df.columns = MultiIndex.from_arrays([a, b, c, d], names=["s", "t", "r", "i"])
    sf = SheetFrame(2, 2, data=df, index=True, sheet=sheet_module)

    a = ["c"] * 10
    b = ["s"] * 5 + ["t"] * 5
    c = ([100] * 2 + [200] * 3) * 2
    x = list(range(10))
    y = list(range(10, 20))
    df = DataFrame({"a": a, "b": b, "c": c, "x": x, "y": y})
    df = df.set_index(["a", "b", "c"])
    sf = SheetFrame(2, 2, data=df, index=True, sheet=sheet_module)

    ax = Axes(left=200, chart_type=ChartType.xlXYScatter)
    x = sf.range("x")
    y = sf.range("y")
    label = sf.first_range("a")
    ax.add_series(x, y, label=label)

    ax.xlabel = sf.range("x", -1)
    ax.ylabel = sf.range("y", -1)

    gr = sf.grouper(None)

    gr = sf.grouper(["a", "b", "c"])
    key = ("c", "t")
    x = gr.range("x", key)
    y = gr.range("y", key)
    label = gr.first_range("b", key)
    ax.add_series(x, y, label=label)

    for x, y in zip(gr.ranges("x"), gr.ranges("y"), strict=True):
        ax.add_series(x, y)

    list(gr.keys())

    list(gr.first_ranges("a"))

    ax.tight_layout()
    ax.set_style()
    ax.set_legend(loc=(0, 0))

    # g = data.groupby("a")
    # key = ("u",)
    # x = data.range("x", g[key])
    # y = data.range("y", g[key])
    # x
    # s = ax.add_series(
    #     x,
    #     y,
    #     # label=data.range("a", g["v"])[0],
    #     chart_type=ChartType.xlXYScatterLines,
    # )
    # ax.xlabel = data.range("x", 0)
    # ax.ylabel = data.range("y", 0)
    # ax.title = "=" + data.range("a", 0).get_address(include_sheetname=True)



import numpy as np
import pytest
from pandas import DataFrame
from xlwings import Sheet

from xlviews.sheetframe import SheetFrame


@pytest.fixture(scope="module")
def sf1(sheet_module: Sheet):
    df = DataFrame({"a": [1, 2, 3, 4], "b": [5, 6, 7, 8]})
    return SheetFrame(2, 3, data=df, style=False, sheet=sheet_module)


@pytest.mark.parametrize(
    ("formula", "value"),
    [
        ("={a}+{b}", [6, 8, 10, 12]),
        ("={a}*{b}", [5, 12, 21, 32]),
        ("={a}-{b}", [-4, -4, -4, -4]),
    ],
)
@pytest.mark.parametrize("use_setitem", [False, True])
def test_formula(sf1: SheetFrame, formula, value, use_setitem):
    if use_setitem:
        sf1["c"] = formula
    else:
        sf1.add_formula_column("c", formula)

    np.testing.assert_array_equal(sf1["c"], value)


@pytest.fixture(scope="module")
def sf2(sheet_module: Sheet):
    df = DataFrame({"a": [1, 2, 3, 4], "b": [5, 6, 7, 8]})
    sf = SheetFrame(10, 3, data=df, style=False, sheet=sheet_module)
    sf.add_wide_column("c", [1, 2, 3, 4])
    return sf


@pytest.mark.parametrize(
    ("formula", "value"),
    [
        ("={a}+{b}+{c}", ([7, 9, 11, 13], [10, 12, 14, 16])),
        ("={a}*{b}*{c}", ([5, 12, 21, 32], [20, 48, 84, 128])),
    ],
)
def test_formula_wide(sf2: SheetFrame, formula, value):
    sf2.add_formula_column("c", formula, number_format="0", autofit=True)
    np.testing.assert_array_equal(sf2[("c", 1)], value[0])
    np.testing.assert_array_equal(sf2[("c", 4)], value[1])



import pytest
from win32com.client import constants


@pytest.fixture(scope="module")
def tlbs():
    from xlviews.client import iter_typelib_specs

    return list(iter_typelib_specs())


@pytest.mark.parametrize(
    "clsid",
    [
        "{00020813-0000-0000-C000-000000000046}",
        "{2DF8D04C-5BFA-101B-BDE5-00AA0044DE52}",
        "{91493440-5A91-11CF-8700-00AA0060263B}",
    ],
)
def test_iter_typelib_specs(tlbs, clsid):
    clsids = [tlb.clsid for tlb in tlbs]
    assert clsid in clsids


def test_ensure_modules():
    from xlviews.client import ensure_modules

    ensure_modules()
    assert constants.msoConnectorStraight == 1



import numpy as np
import pytest
from pandas import DataFrame
from xlwings import Sheet

from xlviews.sheetframe import SheetFrame


@pytest.fixture(scope="module")
def df():
    df = DataFrame(
        {
            "x": ["a"] * 8 + ["b"] * 8 + ["a"] * 4,
            "y": (["c"] * 4 + ["d"] * 4) * 2 + ["c"] * 4,
            "z": range(1, 21),
            "a": range(20),
            "b": list(range(10)) + list(range(0, 30, 3)),
            "c": list(range(20, 40, 2)) + list(range(0, 20, 2)),
        },
    )
    df = df.set_index(["x", "y", "z"])
    df.iloc[[4, -1], 0] = np.nan
    df.iloc[[3, 6, 9], -1] = np.nan
    return df


@pytest.fixture(scope="module")
def sf_parent(df: DataFrame, sheet_module: Sheet):
    return SheetFrame(3, 3, data=df, table=True, sheet=sheet_module)
